#!/usr/bin/env python3

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools, os

with open("README.md", "r") as fh:
    long_description = fh.read()

version   = "0.2." + os.getenv("AZURE_BUILDID", "unknown")

setuptools.setup(
    name="OpenZGY",
    license='Apache',
    version=version,
    #use_scm_version={"root": "..", "relative_to": __file__},
    author="Paal Kvamme",
    author_email="somebody@slb.com", # TODO-High: Change!
    description="Library to access Schlumberger ZGY files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://openzgy.paalsinehoner.net",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=['numpy>=1.17', 'matplotlib', 'pillow', 'zfpy==0.5.5'],
    setup_requires=['setuptools', 'setuptools_scm'],
    # TODO-Low, rename scripts to match command names or vice versa.
    entry_points={
        "console_scripts": [
            "zgydump = openzgy.tools.zgydump:Main",
            "zgycopy = openzgy.tools.copy:Main",
        ],
        "gui_scripts": [
            "zgygui  = openzgy.tools.gui_copy:Main",
        ]
    }
)



#!/usr/bin/env python3

import os
import time
from datetime import datetime
import traceback
import tkinter as tk
import tkinter.filedialog
import tkinter.messagebox
import tkinter.ttk as ttk
from PIL import Image, ImageTk

from ..tools import copy

class TemporaryWidget(tk.Frame):
    def __init__(self, parent=None):
        super().__init__(parent, bd=2)
        self._parent = parent
        self.create_widgets()

    def create_widgets(self):
        if False:
            load = Image.open("dialog.png")
            render = ImageTk.PhotoImage(load)
            self.orig_label = tk.Label(self, image=render)
            self.orig_label.image = render
        elif False:
            self.orig_label = tk.Label(self, text="Compress or decompress ZGY files")
        else:
            self.orig_label = tk.Label(self)
        self.orig_label.grid(sticky="nsew")

class RadioOne(tk.Frame):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._parent = parent
        self.create_widgets()

    def create_widgets(self):
        tk.Radiobutton(self, text="Lossy", padx=0, justify=tk.LEFT,
                       variable=self._parent._parent.compressmode,
                       value=1).grid(row=0, column=1)
        tk.Radiobutton(self, text="Lossless", padx=0, justify=tk.LEFT,
                       variable=self._parent._parent.compressmode,
                       value=2).grid(row=0, column=2)
        tk.Radiobutton(self, text="Uncompressed", padx=0, justify=tk.LEFT,
                       variable=self._parent._parent.compressmode,
                       value=3).grid(row=0, column=3)

class UserInputWidget(tk.Frame):
    def __init__(self, parent=None):
        super().__init__(parent, relief=tk.GROOVE, bd=2)
        self._parent = parent
        self.create_widgets()

    def create_widgets(self):
        self._lbl_inputfile = tk.Label(self, text="Input file:")
        self._lbl_inputfile.grid(row=0, column=0, sticky="w")
        self._txt_inputfile = tk.Entry(self, width=50, textvariable=self._parent.inputname)
        self._txt_inputfile.grid(row=0, column=1, sticky="ew")
        self._btn_inputfile_select = tk.Button(self, text="...", command=self._parent.open_input_dialog)
        self._btn_inputfile_select.grid(row=0, column=2, sticky="w")

        self._lbl_outputfile = tk.Label(self, text="Output file:")
        self._lbl_outputfile.grid(row=1, column=0, sticky="w")
        self._txt_outputfile = tk.Entry(self, width=50, textvariable=self._parent.outputname)
        self._txt_outputfile.grid(row=1, column=1, sticky="ew")
        self._btn_outputfile_select = tk.Button(self, text="...", command=self._parent.open_output_dialog)
        self._btn_outputfile_select.grid(row=1, column=2, sticky="w")

        self._lbl_radio_one = tk.Label(self, text="Output is:")
        self._lbl_radio_one.grid(row=2, column=0, sticky="w")
        self._radio_one = RadioOne(self)
        self._radio_one.grid(row=2, column=1, columnspan=2, sticky="w")

        self._scale_sqnr = tk.Scale(self, orient=tk.HORIZONTAL, from_=20, to=70, tickinterval=10, label="Signal/noise ratio in dB", resolution=5, variable=self._parent.snr)
        self._scale_sqnr.set(30)
        self._scale_sqnr.grid(row=3, column=0, columnspan=3, sticky="eww")

        self._lbl_1 = tk.Label(self, text="<-- Smaller files    Better fidelity -->")
        self._lbl_1.grid(row=4, column=0, columnspan=3, sticky="eww")

        self.grid_columnconfigure(1, weight=1)

class ProgressWidget(tk.Frame):
    def __init__(self, parent=None):
        super().__init__(parent, bd=2)
        self._parent = parent
        self.create_widgets()

    def create_widgets(self):
        self._lbl_last_completed_hdr = tk.Label(self, text="Last completed")
        self._lbl_input_size_hdr = tk.Label(self, text="Input size")
        self._lbl_output_percent_hdr = tk.Label(self, text="New size")
        self._lbl_output_size_hdr = tk.Label(self, text="New size")

        self._lbl_last_completed = tk.Label(self, text="(name)")
        self._lbl_input_size = tk.Label(self, text="XX MB")
        self._lbl_output_percent = tk.Label(self, text="XX %")
        self._lbl_output_size = tk.Label(self, text="XX MB")

        self._lbl_last_completed_hdr.grid(row=0, column=2, padx=5, sticky="w")
        self._lbl_input_size_hdr.grid(row=0, column=3, padx=5)
        self._lbl_output_percent_hdr.grid(row=0, column=4, padx=5)
        self._lbl_output_size_hdr.grid(row=0, column=5, padx=5)

        self._lbl_last_completed.grid(row=1, column=2, padx=5, sticky="w")
        self._lbl_input_size.grid(row=1, column=3, padx=5)
        self._lbl_output_percent.grid(row=1, column=4, padx=5)
        self._lbl_output_size.grid(row=1, column=5, padx=5)

        self._lbl_time_elapsed = tk.Label(self, text="00:00:00", font='TkFixedFont')
        self._lbl_time_elapsed.grid(row=2, column=0, sticky="w")
        self._lbl_time_remain = tk.Label(self, text="00:00:00", font='TkFixedFont')
        self._lbl_time_remain.grid(row=3, column=0)
        self._lbl_elapsed = tk.Label(self, text="elapsed")
        self._lbl_elapsed.grid(row=2, column=1)
        self._lbl_remain = tk.Label(self, text="remain")
        self._lbl_remain.grid(row=3, column=1)
        self._pb_progress = ttk.Progressbar(self)
        self._pb_progress.grid(row=2, column=2, rowspan=2, columnspan=4, sticky="ew")

        self.grid_columnconfigure(2, weight=1)

class BottomWidget(tk.Frame):
    def __init__(self, parent=None):
        super().__init__(parent, bd=2)
        self._parent = parent
        self.create_widgets()

    def create_widgets(self):
        self._estimate = tk.Button(self, text="Estimate size...", state=tk.DISABLED)
        self._estimate.grid(row=0, column=0, padx=5)
        self._run = tk.Button(self, text="Run", command=self._parent.clicked_run)
        self._run.grid(row=0, column=1, padx=5)
        self._stop = tk.Button(self, text="Stop", state=tk.DISABLED, command=self._parent.clicked_stop)
        self._stop.grid(row=0, column=2, padx=5)
        self._quit = tk.Button(self, text="Quit", command=self._parent.master.destroy)
        self._quit.grid(row=0, column=3, padx=5)

class Application(tk.Frame):
    def __init__(self, master=None):
        super().__init__(master)
        self.master = master
        self.grid(sticky="nsew")
        self.create_variables()
        self.create_widgets()
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(3, weight=1)
        self._running = False
        self._stopping = False
        self._has_run = False
        self._start_time = 0
        self._stop_time = 0
        self.update_dependencies()

    def report_callback_exception(self, *args):
        err = traceback.format_exception(*args)
        tkinter.messagebox.showerror("Internal error", err)
        self.master.destroy()

    def set_overwrite_confirmed(self, value):
        """
        True means the output file does not exist or that the user has
        answered "Yes" to the prompt about overwriting. Note that the
        file selection dialog pops up this question itself, so if the
        output file was filled in from there it will already be confirmed.
        Otherwise we wait until the user clicks "Run" before we check.
        """
        #print("overwrite_confirmed(" + str(bool(value)) + ")")
        self.confirmed = value

    def validate(self):
        if not self.inputname.get():
            tkinter.messagebox.showerror("Error", "Input file must be specified.")
            return False
        elif not os.path.exists(self.inputname.get()):
            tkinter.messagebox.showerror("Error", "Input file does not exist.")
            return False

        if not self.outputname.get():
            tkinter.messagebox.showerror("Error", "Output file must be specified.")
            return False
        elif self.outputname.get() == self.inputname.get():
            tkinter.messagebox.showerror("Error", "Input and output cannot be the same file.")
            return False
            # Note that user can sneak around this when typing the name,
            # e.g. foo.zgy and ./foo.zgy.

        elif not self.confirmed:
            if not os.path.exists(self.outputname.get()):
                self.confirmed = True
            else:
                self.confirmed = tkinter.messagebox.askokcancel(
                    "Question",
                    'File "{0}" already exists.\nDo you want to overwrite it?'.format(self.outputname.get()),
                    icon=tkinter.messagebox.WARNING)
        if not self.confirmed:
            return False
        return True

    def update_dependencies(self):
        self.set_state(self._user_input_widget, not self._running)
        self.set_state(self._user_input_widget._scale_sqnr,
                       self.compressmode.get() == 1 and not self._running)
        self.set_state(self._user_input_widget._lbl_1,
                       self.compressmode.get() == 1 and not self._running)
        self.set_state(self._progress_widget, self._running or self._has_run)
        self.set_state(self._bottom_widget._run, not self._running)
        self.set_state(self._bottom_widget._stop, self._running)

    @staticmethod
    def nicenumber(n):
        """
        Format a number with units. Output is always 7 chars.
        """
        if not n: return "       "
        if n < 10000: return "{0:4d}   ".format(n)
        if n < 1024*10000: return "{0:4d} KB".format(n // 1024)
        if n < 1024*1024*10000: return "{0:4d} MB".format(n // (1024*1024))
        return "{0:4d} GB".format(n // (1024*1024*1024))

    def show_results(self):
        isize, osize = (None, None)
        try:
            isize = os.path.getsize(self.inputname.get())
        except Exception:
            pass
        if not self._running and self._has_run:
            try:
                osize = os.path.getsize(self.outputname.get())
            except Exception:
                pass
        isize_str = self.nicenumber(isize) if isize else ""
        osize_str = self.nicenumber(osize) if osize else ""
        osize_pct = "{0:3d} %".format(
            100 * osize // isize) if isize and osize else "     "

        pw = self._progress_widget
        pw._lbl_last_completed_hdr.configure(
            text = "Current file" if self._running else "Last completed")
        shortname = self.inputname.get()
        shortname = os.path.basename(shortname)
        if len(shortname) > 23: shortname = "---" + shortname[-20:]
        pw._lbl_last_completed.configure(text=shortname)
        pw._lbl_input_size.configure(text=isize_str)
        pw._lbl_output_percent.configure(text=osize_pct)
        pw._lbl_output_size.configure(text=osize_str)
        elapsed = "  :  :  "
        elapsed_sec = int(self._stop_time - self._start_time)
        if elapsed_sec > 0 and elapsed_sec < 24*60*60:
            elapsed = datetime.utcfromtimestamp(elapsed_sec).strftime('%H:%M:%S')

        pw._lbl_time_elapsed.configure(text=elapsed)
        pw._lbl_time_remain.configure(text="  :  :  ")
        #self._pb_progress

    def set_state(self, w, state):
        if isinstance(w, tk.Frame):
            for child in w.winfo_children():
                self.set_state(child, state)
        elif isinstance(w, ttk.Progressbar):
            pass # has no "state"
        else:
            s = tk.NORMAL if state else tk.DISABLED
            if False:
                print("Change {0} from {1} to {2}".format(str(w), w.cget('state'), s))
            if isinstance(w, tk.Scale):
                w.configure(state=s, fg="black" if state else "gray")
            w.configure(state=s)

    def set_running(self, on):
        if on and not self._running: self._start_time = time.time()
        if not on and self._running: self._stop_time = time.time()
        self._running = on
        self._stopping = False
        if on: self._has_run = True
        self.update_dependencies()
        self.show_results()

    def clicked_run(self):
        #print("RUN", self)
        if self.validate():
            self.set_running(True)
            mode = self.compressmode.get()
            snr = int(self.snr.get()) if mode==1 else 99 if mode==2 else 0
            cmd = [
                "python3", "-m", "openzgy.tools.copy",
                "--force", "float",
                "--snr", str(snr),
                self.inputname.get(),
                self.outputname.get(),
            ]
            print(*cmd)
            self.set_overwrite_confirmed(False)
            if True:
                copy.copy(
                    self.inputname.get(), self.outputname.get(),
                    progress1=self.update_progress,
                    progress2=lambda done, total: self.update_progress(done, total, True),
                    forcetype='float', snr=snr)
                if self._stopping:
                    # User clicked 'Stop', so the output, if it exists,
                    # is most likely corrupt.
                    self.update_progress(0, 100)
                    if tkinter.messagebox.askokcancel(
                            "Question",
                            'You clicked "Stop",\nso the output file\nis probably unusable.\nDelete it now?', icon=tkinter.messagebox.WARNING):
                        try:
                            os.remove(self.outputname.get())
                        except IOError as ex:
                            tkinter.messagebox.showerror("Error", "While deteling output:\n" + str(ex))
                else:
                    self.update_progress(100, 100, True)
                # If the actual copy is disabled,
                # clicked_stop needs to do this for us.
                self.set_running(False)
        else:
            self.set_running(False)

    def update_progress(self, done, total, flushing=False):
        percent = int(50.0 * done / total)
        if done != 0: percent = max(1, percent)
        if done >= total: percent = 50
        if flushing: percent += 50
        self._progress_widget._pb_progress['value'] = percent
        if False:
            print("DONE:", self._progress_widget._pb_progress['value'],
                  "of", self._progress_widget._pb_progress['maximum'],
                  "flushing LODs" if flushing else "copying")
        root.update()
        return self._running and not self._stopping

    def clicked_stop(self):
        #print("STOP", self)
        self._stopping = True
        # Only when debugging, when the copy process isn't started.
        # Otherwise it will be called when the process finishes.
        #self.set_running(False)

    def showvar(self, name):
        var = getattr(self, name)
        value = var.get()
        if value is str: value = '"' + value + '"'
        print('trace {0} {1} = {2}'.format(name, var, value))

    def create_variables(self):
        self.confirmed = False
        self.inputname = tk.StringVar(self)
        self.outputname = tk.StringVar(self)
        self.outputname.trace("w", lambda *args: self.set_overwrite_confirmed(False))
        self.compressmode = tk.IntVar(self, 1)
        self.compressmode.trace("w", lambda *args: self.update_dependencies())
        self.snr = tk.DoubleVar(self, 30)

        if False: # debug
            self.inputname.trace("w", lambda *args: self.showvar("inputname"))
            self.outputname.trace("w", lambda *args: self.showvar("outputname"))
            self.compressmode.trace("w", lambda *args: self.showvar("compressmode"))
            self.snr.trace("w", lambda *args: self.showvar("snr"))

    def create_widgets(self):
        #self._temporary_widget = TemporaryWidget(self)
        #self._temporary_widget.grid(row=0, column=0, sticky="we")

        self._user_input_widget = UserInputWidget(self)
        self._user_input_widget.grid(row=1, column=0, padx=5, pady=5, sticky="ew")

        self._progress_widget = ProgressWidget(self)
        self._progress_widget.grid(row=2, column=0, sticky="eww")

        self._bottom_widget = BottomWidget(self)
        self._bottom_widget.grid(row=3, column=0, sticky="se", padx=5, pady=5)

    def open_input_dialog(self):
        name = tk.filedialog.askopenfilename(filetypes=[("ZGY", "*.zgy")])
        if name:
            self.inputname.set(name)

    def open_output_dialog(self):
        name = tk.filedialog.asksaveasfilename(filetypes=[("ZGY", "*.zgy")], defaultextension=".zgy")
        if name:
            self.outputname.set(name)
            self.set_overwrite_confirmed(True)

root, app = (None, None)

def Main():
    global root, app
    root = tk.Tk()
    root.title("Compress or decompress ZGY files")
    root.grid_columnconfigure(0, weight=1)
    root.grid_rowconfigure(0, weight=1)
    app = Application(master=root)
    root.report_callback_exception = app.report_callback_exception
    app.mainloop()

if __name__ == "__main__":
    Main()

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

##@package openzgy.impl.genlod

# TODO-Low: Several tweaks are possible but might not offer
# much value. See a detailed discussion in doc/lowres.html.

# Testing notes:
#   test.black.checkLodContents(), checkStatistics(), checkHistogram()
#   is part of a reasonably complete end to end test checking this module.
#   A new unit test testFinalizeProgress() has been written to verify the
#   progress mechanism. Both that the done and total end up precisely
#   identical at the end and that the client can abort the calculation.

import sys
import argparse
import numpy as np
from collections import namedtuple

from . import enum as impl_enum
from ..exception import *
from .stats import StatisticData
from .histogram import HistogramData
from .lodalgo import decimate, DecimationType
from .bulk import ScalarBuffer

def _round_down_to_power_of_2(n):
    exp = -1
    while n > 0:
        n >>= 1
        exp += 1
    return 1<<exp

def _make_blocksizes(bricksize, surveysize, nlods, dtype, factor=(1,1,1), verbose=None):
    """
    CURRENTLY NOT USED.

    Calculate the minimum blocksize to read at each lod level. Clip to
    the survey size. Also compute the memory needed to hold one buffer
    for each lod. Note that the genlod algorithm currently assumes
    that the block size is the same for all levels except for the
    clipping. And it currently handles clipping itself and might not
    like us to do so. Currently this function is not very useful.
    """
    blocksizes = np.zeros((nlods, 3), dtype=np.int64)
    ss = np.array(surveysize, dtype=np.int64)
    bs = np.array([2*factor[0]*bricksize[0],
                   2*factor[1]*bricksize[1],
                   ss[2]], dtype=np.int64)
    iterations = 0
    for lod in range(nlods):
        bs = np.minimum(bs, ss)
        blocksizes[lod] = bs
        iterations += np.product((ss+bs-1) // bs)
        ss = (ss + 1) // 2
    bytesused = np.sum(np.product(blocksizes, axis=1)) * int(np.dtype(dtype).itemsize)
    returntype = namedtuple("BlockSizeInfo", "blocksizes bytesused iterations")
    result = returntype(blocksizes, bytesused, iterations)
    print(result)
    return result

def _choose_blocksizes(bricksize, surveysize, nlods, dtype, maxmem=512*1024*1024):
    """
    CURRENTLY NOT USED.

    Plan B and C are rather memory intensive already. So we might not
    be able to increase the buffer sizes very much.

    Calculate the actual blocksize to read at each lod level. If there
    is enough memory available we can multiply the basic block size by
    some factor in the j direction. And if there is still more memory
    left it can be multiplied in the J directon as well. There is a
    benefit when the file resides on the cloud because the code will
    read larger blocks from lod 0. There is also a benefit both for
    cloud and on-prem files in that low resolutun bricks end up more
    contiguous.

    Beware that the current genlod algorithm assumes the block size is
    the same at each level. It will do exactly 4 reads at lod-1 to
    produce one brick at lod. There is also a requirement that the
    facors are a power of 2. Neither of those limitations apply to the
    non-recursive plan B. Which is currently not implemented.

    This means that the factor we multipy with should be the same for
    all lods. Had it not been for this, it would probably have made
    sense to make the blocks larger in (i,j) since they will be
    smaller in k.

    For very small surveys there is another caveat. The granularity
    of calls to the progress callback will be the block size.
    If the buffer gets too large then a progress bar won't be
    very useful.
    """
    result  = _make_blocksizes(bricksize, surveysize, nlods, dtype,
                               factor=(1,1,1))
    jfactor = max(1, maxmem // result.bytesused)
    jfactor = _round_down_to_power_of_2(jfactor)
    result  = _make_blocksizes(bricksize, surveysize, nlods, dtype,
                               factor=(1,jfactor,1))
    ifactor = max(1, maxmem // result.bytesused)
    ifactor = _round_down_to_power_of_2(ifactor)
    result = _make_blocksizes(bricksize, surveysize, nlods, dtype,
                              factor=(ifactor,jfactor,1), verbose=None)
    return result

class GenLodBase:
    """
    Abstract class for generating low resolution bricks, histogram,
    and statistics. At this level only define virtual methods for I/O.
    The implementation can be used as-is when mocking the class.
    The optional nlods parameter is only used as a consistency check.

    Note that the WeightedAverage algorithm requires a histogram to work.
    If no histogram was provided then the current contents of the
    accumulated histogram will be used. This is unfortunate and might
    cause brick artifacts. Especially in the first few bricks that are
    generated. With a non-recursive algorithm (plan C) and with only
    lod2 and above uses weighted average then this is unproblematic.
    Because in that case we will be done with the histogram once we
    need it. TODO-Low consider doing an initial pass with a statistical
    sampling of the lod0 data, only for use with weighted average.
    There will be a minor issue with some values appearing to have zero
    frequency, but this should not cause any trouble. (assume "1").

    Note that the WeightedAverage and AverageNon0 algorithms expect a
    defaultvalue to use when all inputs are inf/nan or (for AverageNon0)
    zero. Only relevant for integral types, to ensure that the default
    is whatever will produce the value closest to 0 after conversion.
    And integral data can neither be inf nor nan, so this is a pretty
    academic issue. For AverageNon0 that algorithm is currently not
    used. So it isn't even clear what the desired behavior should be.
    """
    def __init__(self, size, *, bricksize = (64, 64, 64), dtype = np.float32,
                 range_hint = None, nlods = None,
                 decimation = None, histogram = None, defaultvalue = None,
                 progress = None, verbose = None):
        _nlods = 1 # The loop stops before counting final level
        _total = 1 # Total number of bricks in all levels.
        bs = np.array(bricksize, dtype=np.int64)
        sz = np.array(size, dtype=np.int64)
        while np.any(sz > bs):
            _nlods += 1
            _total += np.product((sz + bs - 1) // bs)
            sz = (sz + 1) // 2
        assert nlods is None or nlods == _nlods
        self._surveysize = np.array(size, dtype=np.int64)
        self._bricksize = np.array(bricksize, dtype=np.int64)
        self._dtype = dtype
        self._range_hint = range_hint or (-1, 1)
        self._progress = progress
        self._verbose = verbose or (lambda *args, **kw: False)
        self._done = 0
        self._total = _total
        self._nlods = _nlods
        self._surveysize.flags.writeable = False
        self._bricksize.flags.writeable = False
        self._decimation_type = decimation or [DecimationType.LowPass,
                                               DecimationType.WeightedAverage]
        self._wa_histogram = histogram # Might become self._histo later.
        self._wa_defaultvalue = defaultvalue or 0
        self._verbose("@survey {0}".format(tuple(self._surveysize)))
        #_choose_blocksizes(bricksize, size, nlods, dtype)
        self._report(None)

    def _report(self, data):
        """
        Invoke the user's progress callback if any.
        Keep track of how many bricks we have processed. Both reads and
        writes increment the same counter. For plan C the reads will cover
        all blocks in lod 0 and the writes will cover all blocks in lod > 0.
        For plan D all blocks are written which means the computation of
        _total done in __init__ might need to change.
        """
        if data is not None:
            count = np.product((np.array(data.shape, dtype=np.int64) +
                                self._bricksize - 1) // self._bricksize)
            self._done += count
        if self._progress and not self._progress(self._done, self._total):
            raise ZgyAborted("Computation of low resolution data was aborted")

    def _read(self, lod, pos, size):
        """
        This is a stub that must be redefined except for low level unit tests.
        Read a block from the ZGY file (plans B and C) or the application
        (plan D). The lod parameter will always be 0 for plans C and D.
        Returns a ScalarBuffer if all constant, else a 3D numpy array.
        """
        result = ScalarBuffer(size, 0, self._dtype)
        self._report(result)
        return result

    def _write(self, lod, pos, data):
        """
        This is a stub that must be redefined except for low level unit tests.
        Write a block to the ZGY file. Silently ignore writes of data that
        is known to have been read directly from the file. For plans B and C
        this means ignoring all writes to lod 0.
        """
        self._report(data)

    def _savestats(self):
        """
        This is a stub that must be redefined except for low level unit tests.
        Finalize and write the computed statistics and histogram to the file.
        """
        pass

    def _prefix(self, lod):
        """For debugging and logging only."""
        return "  " * (self._nlods-1 - lod)

    @staticmethod
    def _format_result(result):
        """For debugging and logging only."""
        if isinstance(result, np.ndarray):
            return "array" + str(tuple(result.shape))
        else:
            return repr(result)

class GenLodImpl(GenLodBase):
    """
    Abstract class for generating low resolution bricks, histogram,
    and statistics. The inherited methods for I/O are still stubs.
    See doc/lowres.html for details. This class implements plan C or D
    which is good for compressed data and acceptable for uncompressed.
    The ordering of low resolution bricks in the file will not be optimal.
    For optimal ordering but working only for uncompressed data consider
    implementing plan B in addition to the plan C already implemented.
    The implementation can be used as-is in a unit test with mocked I/O.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        #print("GenLodImpl", args, kwargs)
        self._stats = StatisticData()
        self._histo = HistogramData(range_hint=self._range_hint,
                                    dtype=self._dtype)
        # See base constructor. Better than no histogram at all.
        self._wa_histogram = self._wa_histogram or self._histo

    def __call__(self):
        """
        Generate and store statistics, histogram, and all low resolution
        bricks. Works for plans C and D. If we also need an implementation
        of plan B then this method wold need to iterate over all bricks
        and lods, and _accumulate would not make any recursive calls.
        """
        self._calculate((0,0,0), self._nlods-1)
        self._savestats() # Remove?
        return self._stats, self._histo

    def _accumulate(self, data):
        """
        Keep a running tally of statistics and histogram.
        """
        if data is None:
            return
        factor = 1
        if isinstance(data, ScalarBuffer):
            factor = np.product(data.shape, dtype=np.int64)
            data = np.array([data.value], dtype=data.dtype)
        self._stats.add(data, factor)
        self._histo.add(data, factor)

    def _calculate(self, readpos, readlod):
        """
        Read data from the specified (readpos, readlod) and store it back.
        The function will itself decide how much to read. But with several
        constraints. Always read full traces. Size in i and j needs to be
        2* bs * 2^N  where bs is the file's brick size in that dimension,
        Clipped to the survey boundaries. This might give an empty result.

        When readlod is 0 and the data was read from the ZGY file then the
        writing part is skipped. Since the data is obviously there already.

        In addition to reading and writing at the readlod level, the
        method will compute a single decimated buffer at readlod+1 and
        return it. As with the read/write the buffer might be smaller at
        the survey edge. Note that the caller is responsible for storing
        the decimated data.

        Full resolution data (lod 0) will be read from file (plan C) or the
        application (plan D). Low resolution is computed by a recursive call
        to this function (plans C and D) or by reading the file (plan B).
        Note that currently only plan C is fully implemented.

        For plans B and C a single call needs to be made to read the brick
        (there is by definition just one) at the highest level of detail.
        This will end up computing all possible low resolution bricks and
        storing them. For plan B the caller must iterate.
        """
        surveysize = (self._surveysize + (1<<readlod) - 1) // (1<<readlod)
        readpos = np.array((readpos[0], readpos[1], 0), dtype=np.int64)

        if readpos[0] >= surveysize[0] or readpos[1] >= surveysize[1]:
            return None

        readsize = np.minimum(2*self._bricksize, surveysize - readpos)
        readsize[2] = surveysize[2] # Always read full traces.
        writesize = (readsize + 1) // 2

        self._verbose("@{0}calculate(lod={1}, pos={2})".format(
            self._prefix(readlod), readlod, tuple(readpos)))
        if readlod == 0:
            data = self._read(lod=readlod, pos=readpos, size=readsize)
            self._accumulate(data)
        else:
            offsets = [[0, 0, 0], [0, 1, 0], [1, 0, 0], [1, 1, 0]]
            offsets = np.array(offsets, dtype=np.int64) * self._bricksize
            hires = [None, None, None, None]
            for i in range(4):
                hires[i] = self._calculate(2*readpos + 2*offsets[i], readlod-1)
            data = self._paste4(hires[0], hires[1], hires[2], hires[3])

        self._write(lod=readlod, pos=readpos, data=data)

        if readlod == self._nlods - 1:
            result = None # Caller will discard it anyway.
            assert self._done == self._total
        elif not isinstance(data, np.ndarray):
            result = ScalarBuffer(writesize, data.value, data.dtype)
        else:
            result = self._decimate(data, lod=readlod+1)
            assert tuple(result.shape) == tuple(writesize)

        self._verbose("@{0}calculate returns(lod={1}, pos={2}, data={3})".format(
            self._prefix(readlod), readlod+1, tuple(readpos//2),
            self._format_result(result)))
        return result

    def _decimate(self, data, lod):
        """
        Return a decimated version of the input buffer with half the size
        (rounded up) in each dimension. In total the result will be ~1/8
        the size of the input.
        """
        if data is None:
            return None
        elif isinstance(data, ScalarBuffer):
            size = ((data.shape[0]+1)//2,
                    (data.shape[1]+1)//2,
                    (data.shape[2]+1)//2)
            return ScalarBuffer(size, data.value, data.dtype)
        else:
            di = data.shape[0] % 2
            dj = data.shape[1] % 2
            dk = data.shape[2] % 2
            if di != 0 or dj != 0 or dk != 0:
                data = np.pad(data, ((0, di), (0, dj), (0, dk)), mode='edge')
            #return data[::2,::2,::2]
            dt = self._decimation_type[min(lod, len(self._decimation_type)-1)]
            if dt == DecimationType.WeightedAverage:
                return decimate(data, dt,
                                histogram = self._wa_histogram,
                                defaultvalue = self._wa_defaultvalue)
            elif dt == DecimationType.AverageNon0:
                return decimate(data, dt,
                                defaultvalue = self._wa_defaultvalue)
            else:
                return decimate(data, dt)

    def _paste1(self, result, more, ioff, joff):
        """
        See _paste4() for details.
        """
        if more is not None:
            value = more.value if isinstance(more, ScalarBuffer) else more
            result[ioff:ioff+more.shape[0], joff:joff+more.shape[1], :] = value

    def _paste4(self, d00, d01, d10, d11):
        """
        Combine 4 buffers into one. Input buffers may be None (do not
        paste) or ScalarBuffer (paste a constant value). If all not-None
        buffers are just scalars then the return from this function
        will also be a scalar. d01 adds more data in the J direction,
        so it starts at i=0, j>0 in the target. Similarly d10 adds
        more in the J direction. And d11 in the diagonal.
        """
        if d01 is None and d10 is None and d11 is None:
            return d00 # Nothing to paste. Also works for None or scalar.

        assert d00 is not None # Empty d00 + non-empty others is not good.
        assert d01 is None or d01.shape[0] == d00.shape[0]
        assert d10 is None or d10.shape[1] == d00.shape[1]
        if d01 is not None and d10 is not None:
            # The "diagonal" brick must exist with the right size.
            assert d11 is not None
            assert d11.shape[1] == d01.shape[1] and d11.shape[0] == d10.shape[0]
        else:
            # The "diagonal" brick should not be needed in this case.
            assert d11 is None

        ni = d00.shape[0] + (d10.shape[0] if d10 is not None else 0)
        nj = d00.shape[1] + (d01.shape[1] if d01 is not None else 0)
        nk = d00.shape[2]

        all_same = True
        for e in (d00, d01, d10, d11):
            if all_same and e is not None:
                if not isinstance(e, ScalarBuffer) or e.value != d00.value:
                    all_same = False
        if all_same:
            result = ScalarBuffer((ni, nj, nk), d00.value, d00.dtype)
        else:
            result = np.zeros((ni, nj, nk), dtype=d00.dtype)
            self._paste1(result, d00, 0, 0)
            self._paste1(result, d01, 0, d00.shape[1])
            self._paste1(result, d10, d00.shape[0], 0)
            self._paste1(result, d11, d00.shape[0], d00.shape[1])
        return result

class GenLodC(GenLodImpl):
    """
    Generate and store low resolution bricks, histogram, and statistics.
    See doc/lowres.html for details. I/O is done via ZgyInternalBulk.
    Use this class as part as finalize().
    The implementation uses plan C, which means the full resolution data
    will be read from the ZGY file. To implement plan D, make a derived
    class that redefines _read() to query the client for the required full
    resolution data. _read() must then also call _write() to store the
    data it just received.
    """
    def __init__(self, accessor, *,
                 compressor = None, decimation = None,
                 progress = None, verbose = None):
        # Could add defaultvalue. And add a unit test for correct operation.
        super().__init__(size = accessor._metadata._ih._size,
                         bricksize = accessor._metadata._ih._bricksize,
                         dtype = impl_enum._map_DataTypeToNumpyType(
                             accessor._metadata._ih._datatype),
                         range_hint = (accessor._sample_min,
                                       accessor._sample_max),
                         nlods = accessor._metadata._ih._nlods,
                         decimation = decimation,
                         progress = progress, verbose = verbose)
        self._accessor = accessor
        self._compressor = compressor

    def _read(self, lod, pos, size):
        """See base class for detils."""
        as_float = bool(self._dtype == np.float32)
        verbose  = False
        data = self._accessor.readConstantValue(
            pos, size, lod=lod, as_float=as_float, verbose=verbose)
        if data is not None:
            data = ScalarBuffer(size, data, self._dtype)
        else:
            data = np.zeros(size, dtype=self._dtype)
            self._accessor.readToExistingBuffer(
                data, pos, lod=lod, as_float=as_float, verbose=verbose)
        self._verbose("@{0}read(lod={1}, pos={2}, size={3}) => {4}".format(
            self._prefix(lod-1), lod, tuple(pos), tuple(size),
            self._format_result(data)))
        assert tuple(size) == tuple(data.shape)
        self._report(data)
        return data

    def _write(self, lod, pos, data):
        """See base class for detils."""
        self._verbose("@{0}write(lod={1}, pos={2}, data={3})".format(
            self._prefix(lod-1), lod, tuple(pos),
            self._format_result(data)))
        is_storage = bool(data.dtype != np.float32)
        verbose  = False
        if lod > 0:
            self._accessor._writeRegion(
                data, pos, lod=lod, compressor=self._compressor,
                is_storage=is_storage, verbose=verbose)
            self._report(data)

    def _savestats(self):
        #print("STATS", repr(self._stats), str(type(self._stats._min)))
        pass

def main(filename, snr):
    """
    Create or re-create all statistics, histogram, and low resolution data
    in the provided open ZGY file. This method is for testing only; normally
    class GenLodC will be invoked from inside the api layer as part of
    finalizing a newly created file. Normally the genlod module should not
    use the api layer itself. Also note that ZGY technically doesn't allow
    updating a file once it has been written and closed. A kludge makes
    it work, sort of, but not for seismic store and not for versions < 2.
    Also, only the lowres data is actually written to the file.
    Now that the new code has been hooked up to the rest of OpenZGY there
    isn't really much need for this stand alone test; the code should be
    covered better by the regular unit tests.
    """
    from ..api import ZgyReader, ZgyCompressFactory, ProgressWithDots
    from ..test.utils import SDCredentials
    with ZgyReader(filename, iocontext=SDCredentials(), _update=True) as reader:
        GenLodC(accessor   = reader._accessor,
                progress   = ProgressWithDots(),
                verbose    = lambda *args, **kw: None, # print(*args, **kw),
                compressor = ZgyCompressFactory("ZFP", snr = snr))()

if __name__ == "__main__":
    """
    Stand alone app for testing the genlod module. See main() for details.
    """
    np.seterr(all='raise')
    parser = argparse.ArgumentParser(description="Generate and store low resolution bricks, histogram, and statistics. See doc/lowres.html for details.")
    parser.add_argument('input', help='ZGY input cube, local or sd://')
    parser.add_argument('--snr', default=0, type=int,
                        help='Pass 10..70 for lossy compression, 99 for lossless, omit for uncompressed.')
    args = parser.parse_args()
    if not args.input:
        print("File names cannot be empty.", file=sys.stderr)
        sys.exit(1)
    main(args.input, args.snr)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

#print('Running' if __name__ == '__main__' else 'Importing', __file__)

import numpy as np
import sys
import os
import argparse
from PIL import Image
import tkinter as tk

from ..api import ZgyReader
from ..test.utils import SDCredentials
from ..tools.viewzgy import savePNG, showFileInTk

def run(filename, *, lods = [0], direction = 0, datarange=None,
        outname = None, iocontext=None):
    """
    Read one trace or slices in the specified direction, save and display.
    """
    allslices = []
    with ZgyReader(filename, iocontext=SDCredentials()) as reader:
        for lod in lods:
            step = 1<<lod
            start = [0, 0, 0]
            size = np.array(reader.size, dtype=np.int64) // (1 << lod)
            start[direction] = size[direction] // 2
            size[direction] = 1
            if direction == 2:
                size[0] = min(size[0], 1024)
                size[1] = min(size[1], 1024)
            section = np.zeros(size, dtype=np.float32)
            reader.read(start, section, lod)
            allslices.append(np.squeeze(section))
    s = allslices[0].shape
    w = np.sum([allslices[i].shape[0] for i in range(3)])
    combined = np.zeros((w, s[1]), dtype=allslices[0].dtype)
    combined[0:s[0],:] = allslices[0]
    ss = allslices[1].shape
    combined[s[0]:s[0]+ss[0], 0:ss[1]] = allslices[1]
    sss = allslices[2].shape
    combined[s[0]+ss[0]:s[0]+ss[0]+sss[0], 0:sss[1]] = allslices[2]
    combined = combined[::-1,::]
    savePNG(combined, outname, datarange=datarange)
    showFileInTk(outname, title=os.path.basename(outname))

def Main(args, *, gain="1", prefix=""):
    """
    Show 3 ZGY files, with the 3 first LODs.
    The files are assumed to be uncompressed original, file after compression,
    and the difference between those two. The value range of the diff is a
    certain percentage of the value range in the first inout. This makes it
    possible to compare the quality from different settings.
    """
    with ZgyReader(args[0], iocontext=None) as reader:
        datarange0 = reader.datarange
        datarange1 = ((reader.datarange[0]/gain, reader.datarange[1]/gain))
    run(args[0], lods=range(3), direction=0, datarange = datarange0,
            outname = prefix + "-showorig.png")
    run(args[1], lods=range(3), direction=0, datarange = datarange0,
            outname = prefix + "-showcomp.png")
    run(args[2], lods=range(3), direction=0, datarange = datarange1,
            outname = prefix + "-showdiff.png")

if __name__ == "__main__":
    np.seterr(all='raise')
    parser = argparse.ArgumentParser(description='Evaluate compression')
    parser.add_argument('orig', help='Original ZGY input file')
    parser.add_argument('comp', help='Data after compression')
    parser.add_argument('diff', help='Compression noise')
    parser.add_argument('--gain', nargs=1, default=[50], type=int,
                        help='Scaling of diff cube relative to orig range.')
    parser.add_argument('--prefix', nargs=1, default=["check"], type=str,
                        help='prefix for output file name.')
    args = parser.parse_args()
    #print(args)
    Main([args.orig, args.comp, args.diff], gain=args.gain[0], prefix=args.prefix[0])
    sys.exit(0)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Consolidate old and new Python API.

This file is for testing only, when comparing the old and new apis.
The Python wrapper for the old ZGY-Public API is not quite identical
to the new OpenZGY api. This file wraps the old API to become more
similar to the new.

Application code should NOT use this package when migrating. Instead,
please modify the code to fit the slightly changed API. The documentation
for this package might help to show which changes are required.
"""
##@package openzgy.zgypublic
#@brief Consolidate old and new Python API.

import zgy
from .api import SampleDataType, UnitDimension

def _fixKeywordArgs(kwargs):
    """
    The new API uses real enums instead of stringly typed values.
    The names of each tag matches, so the conversion is trivial.
    Also handles "iocontext" and "sourcetype" although those two
    are not in either interface yet.
    """
    result = {}
    for key, value in kwargs.items():
        if key in ("zunitdim", "hunitdim", "datatype", "sourcetype"):
            result[key] = value.name
        elif key in ("iocontext", "verbose"):
            pass
        else:
            result[key] = value
    return result

def _fixResultData(kwdata):
    """
    Inverse of _fixKeywordArgs. Convert strings to enums.
    """
    result = {}
    for key, value in kwdata.items():
        if key in ("datatype", "sourcetype"):
            result[key] = SampleDataType[value]
        elif key in ("zunitdim", "hunitdim"):
            result[key] = UnitDimension[value]
        else:
            result[key] = value
    return result

# Note, using mostly Doxygen style annotation because for this file
# it isn't really useful to have help instde Python.

class ZgyReader(zgy.ZgyReader):
    ##@brief Ignore iocontext and verbose arguments.
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **_fixKeywordArgs(kwargs))

    ##@brief Ignore iocontext and verbose arguments.
    def read(self, *args, **kwargs):
        return super().read(*args, **_fixKeywordArgs(kwargs))

    ##@brief Old: strings, new: enums for 4 properties.
    #@details datatype, sourcetype, zunitdim, hunitdim need to be mapped.
    @property
    def meta(self):
        return _fixResultData(super().meta)

    ##@brief Old: string, new: openzgy.api.SampleDataType.
    @property
    def datatype(self):
        return SampleDataType[super().datatype]

    ##@brief Attribute doesn't exist in old api
    @property
    def raw_datarange(self):
        return super().datarange

    ##@brief Old: string, new: openzgy.api.UnitDimension.
    @property
    def zunitdim(self):
        return UnitDimension[super().zunitdim]

    ##@brief Old: string, new: openzgy.api.UnitDimension.
    @property
    def hunitdim(self):
        return UnitDimension[super().hunitdim]

    ##@brief Old: missing from api.
    #@details For the old API the bricksize will always be returned as
    #(64,64,64) on read and always set to that value on file create.
    @property
    def bricksize(self):
        return (64, 64, 64)

class ZgyWriter(zgy.ZgyWriter):
    ##@brief Ignore iocontext and verbose, and map 4 enums to string.
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **_fixKeywordArgs(kwargs))

    ##@brief Ignore iocontext and verbose arguments.
    def write(self, *args, **kwargs):
        return super().write(*args, **_fixKeywordArgs(kwargs))

    ##@brief Ignore and let _\_exit_\_ handle it instead.
    def finalize(*args, **kwargs):
        return None

    ##@brief Old: strings, new: enums for 4 properties.
    #@details datatype, sourcetype, zunitdim, hunitdim need to be mapped.
    @property
    def meta(self):
        return _fixResultData(super().meta)

    ##@brief Old: string, new: openzgy.api.SampleDataType.
    @property
    def datatype(self):
        return SampleDataType[super().datatype]

    ##@brief Attribute doesn't exist in old api
    @property
    def raw_datarange(self):
        return super().datarange

    ##@brief Old: string, new: openzgy.api.UnitDimension.
    @property
    def zunitdim(self):
        return UnitDimension[super().zunitdim]

    ##@brief Old: string, new: openzgy.api.UnitDimension.
    @property
    def hunitdim(self):
        return UnitDimension[super().hunitdim]

    ##@brief Old: missing from api.
    #@details For the old API the bricksize will always be returned as
    #(64,64,64) on read and always set to that value on file create.
    @property
    def bricksize(self):
        return (64, 64, 64)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



# pylint: disable=invalid-name,no-member
"""
Basic definitions and tools for the creation of seismic datasets. This is an
xarray.Dataset with particular dimensions, coordinates and attributes that inform
other tools and file format converters in segysak library.

"""

import numpy as np

from ._keyfield import (
    CoordKeyField,
    VerticalUnits,
    AttrKeyField,
    VerticalKeyField,
    DimensionKeyField,
)
from ._accessor import xr


def _check_input(var):
    if var is None:
        return var
    if isinstance(var, int):
        return np.array([i for i in range(1, var + 1)], dtype=int)
    else:
        var = np.asarray(var)
        if var.ndim > 1:
            raise ValueError("Inputs for geometry must be one dimensional")
        return var


def create_seismic_dataset(
    twt=None, depth=None, cdp=None, iline=None, xline=None, offset=None, **dim_args
):
    """Create a blank seismic dataset by setting the dimension sizes (d#) or by passing
    arrays for known dimensions.

    iline and xline must be specified together and are mutually exclusive to cdp argument.

    Args:
        twt (int/array-like, optional): Two-way time vertical sampling coordinates.
            Cannot be used with depth argument. Defaults to None.
        depth (int/array-like, optional): Depth vertical sampling coordinates.
            Cannot be used with twt argument. Defaults to None.
        cdp (int/array-like, optional): The CDP numbering for 2D data, cannot be used
            with iline or xline. Use for 2D seismic data. Defaults to None.
        iline (int/array-like, optional): The iline numbering, cannot be
            used with cdp argument. Use for 3D seismic data. Defaults to None.
        xline (int/array-like, optional): The xline numbering, cannot be
            used with cdp argument. Use for 3D seismic data. Defaults to None.
        offset (int/array-like, optional): The offset. This will fill dimension d4.
            Use for pre-stack data. Defaults to None.
        dim_args (int/array-like): Other dimensions you would like in your dataset. The key will be the dimension name.

    Returns:
        xarray.Dataset: A dataset with the defined dimensions of input setup to work with seisnc standards.
    """
    # cdp not allowed with iline or xline
    if cdp is not None and (iline is not None or xline is not None):
        raise ValueError(
            "cdp argument cannot be used with 3D dimensions (iline or xline)"
        )

    if iline is not None and xline is None:
        raise ValueError("xline needed with iline argument, 3d geometry requires both")

    if xline is not None and iline is None:
        raise ValueError("xline needed with iline argument, 3d geometry requires both")

    # check inputs
    twt = _check_input(twt)
    depth = _check_input(depth)
    cdp = _check_input(cdp)
    iline = _check_input(iline)
    xline = _check_input(xline)
    offset = _check_input(offset)
    # # make sure other dim args are sensible and to spec
    for key in dim_args.keys():
        dim_args[key] = _check_input(dim_args[key])

    dimensions = dict()
    # create dimension d1
    if cdp is not None:
        dimensions[CoordKeyField.cdp] = ([CoordKeyField.cdp], cdp)
    elif iline is not None:  # 3d data
        dimensions[CoordKeyField.iline] = ([CoordKeyField.iline], iline)
        dimensions[CoordKeyField.xline] = ([CoordKeyField.xline], xline)

    # create dimension d3
    if twt is not None:
        dimensions[CoordKeyField.twt] = ([CoordKeyField.twt], twt)
    if depth is not None:
        dimensions[CoordKeyField.depth] = ([CoordKeyField.depth], depth)

    # create dimension d4
    if offset is not None:
        dimensions[CoordKeyField.offset] = (
            [CoordKeyField.offset],
            offset,
        )

    # any left over dims
    for arg, val in dim_args.items():
        dimensions[arg] = ([arg], val)

    ds = xr.Dataset(coords=dimensions)

    if twt is not None:
        ds.attrs[AttrKeyField.ns] = twt.size
    elif depth is not None:
        ds.attrs[AttrKeyField.ns] = depth.size

    ds.attrs.update({name: None for name in AttrKeyField})

    return ds


def _check_vert_units(units):
    # check units
    if units is not None:
        try:
            units = VerticalUnits[units]
        except (AttributeError, KeyError):
            raise ValueError(
                f"The vert_units is unknown, got {units}, expected one of {VerticalUnits.keys()}"
            )
    return units


def _dataset_coordinate_helper(vert, vert_domain, **kwargs):
    if vert_domain == "TWT":
        kwargs["twt"] = vert
        domain = VerticalKeyField.twt
    elif vert_domain == "DEPTH":
        kwargs["depth"] = vert
        domain = VerticalKeyField.depth
    else:
        raise ValueError(
            f"Unknown vert_domain {vert_domain}, expected one of TWT or DEPTH"
        )
    return kwargs, domain


def create3d_dataset(
    dims,
    first_sample=0,
    sample_rate=1,
    first_iline=1,
    iline_step=1,
    first_xline=1,
    xline_step=1,
    first_offset=None,
    offset_step=None,
    vert_domain="TWT",
    vert_units=None,
):
    """Create a regular 3D seismic dataset from basic grid geometry with optional
    offset dimension for pre-stack data.

    Args:
        dims (tuple of int): The dimensions of the dataset to create (iline, xline, vertical).
            If first_offset is specified then (iline, xline, vertical, offset)
        first_sample (int, optional): The first vertical sample. Defaults to 0.
        sample_rate (int, optional): The vertical sample rate. Defaults to 1.
        first_iline (int, optional): First inline number. Defaults to 1.
        iline_step (int, optional): Inline increment. Defaults to 1.
        first_xline (int, optional): First crossline number. Defaults to 1.
        xline_step (int, optional): Crossline increment. Defaults to 1.
        first_offset (int/float, optional): If not none, the offset dimension will be added starting
            at first offset. Defaults to None.
        offset_step (int, float, optional): Required if first_offset is specified. The offset increment.
        vert_domain (str, optional): Vertical domain, one of ('DEPTH', 'TWT'). Defaults to 'TWT'.
        vert_units(str, optional): Measurement system of of vertical coordinates.
            One of ('ms', 's', 'm', 'km', 'ft'): Defaults to None for unknown.
    """
    vert_domain = vert_domain.upper()

    if first_offset is None:
        ni, nx, ns = dims
    else:
        ni, nx, ns, no = dims

    units = _check_vert_units(vert_units)

    # 1e-10 to stabilise range on rounding errors for weird floats from hypothesis

    vert = np.arange(
        first_sample,
        first_sample + sample_rate * ns - 1e-10,
        sample_rate,
        dtype=int,
    )
    ilines = np.arange(
        first_iline,
        first_iline + iline_step * ni - 1e-10,
        iline_step,
        dtype=int,
    )
    xlines = np.arange(
        first_xline,
        first_xline + xline_step * nx - 1e-10,
        xline_step,
        dtype=int,
    )

    if first_offset is None:
        offset = None
    else:
        offset = np.arange(
            first_offset,
            first_offset + offset_step * no - 1e-10,
            offset_step,
            dtype=float,
        )

    builder, domain = _dataset_coordinate_helper(
        vert, vert_domain, iline=ilines, xline=xlines, offset=offset
    )

    ds = create_seismic_dataset(**builder)
    ds.attrs[AttrKeyField.measurement_system] = units
    ds.attrs[AttrKeyField.d3_domain] = domain
    ds.attrs[AttrKeyField.sample_rate] = sample_rate
    ds.attrs[AttrKeyField.text] = "SEGY-SAK Create 3D Dataset"
    ds.attrs[AttrKeyField.corner_points] = [
        (ilines[0], xlines[0]),
        (ilines[-1], xlines[0]),
        (ilines[-1], xlines[-1]),
        (ilines[0], xlines[-1]),
        (ilines[0], xlines[0]),
    ]

    return ds


def create2d_dataset(
    dims,
    first_sample=0,
    sample_rate=1,
    first_cdp=1,
    cdp_step=1,
    first_offset=None,
    offset_step=None,
    vert_domain="TWT",
    vert_units=None,
):
    """Create a regular 2D seismic dataset from basic geometry.

    Args:
        dims (tuple of int): The dimensions of the dataset to create (ncdp, vertical).
            If first_offset is specified then (ncdp, vertical, offset)
        first_sample (int, optional): The first vertical sample. Defaults to 0.
        sample_rate (int, optional): The vertical sample rate. Defaults to 1.
        first_cdp (int, optional): First CDP number. Defaults to 1.
        cdp_step (int, optional): CDP increment. Defaults to 1.
        first_offset (int/float, optional): If not none, the offset dimension will be added starting
            at first offset. Defaults to None.
        offset_step (int, float, optional): Required if first_offset is specified. The offset increment.
        vert_domain (str, optional): Vertical domain, one of ('DEPTH', 'TWT'). Defaults to 'TWT'.
        vert_units(str, optional): Measurement system of of vertical coordinates.
            One of ('ms', 's', 'm', 'km', 'ft'): Defaults to None for unknown.
    """
    vert_domain = vert_domain.upper()

    if first_offset is None:
        ncdp, ns = dims
    else:
        ncdp, ns, no = dims

    # check units
    units = _check_vert_units(vert_units)

    vert = np.arange(
        first_sample, first_sample + sample_rate * ns, sample_rate, dtype=int
    )
    cdps = np.arange(first_cdp, first_cdp + cdp_step * ncdp, cdp_step, dtype=int)

    if first_offset is None:
        offset = None
    else:
        offset = np.arange(
            first_offset, first_offset + offset_step * no, offset_step, dtype=float
        )

    builder, domain = _dataset_coordinate_helper(
        vert, vert_domain, cdp=cdps, offset=offset
    )

    ds = create_seismic_dataset(**builder)
    ds.attrs[AttrKeyField.measurement_system] = units
    ds.attrs[AttrKeyField.d3_domain] = domain
    ds.attrs[AttrKeyField.sample_rate] = sample_rate
    ds.attrs[AttrKeyField.text] = "SEGY-SAK Create 2D Dataset"

    return ds



import xarray as xr
import segyio
import numpy as np
import pandas as pd
from more_itertools import split_into, split_when
from itertools import product

from ._segy_core import tqdm, check_tracefield

from .._accessor import open_seisnc
from .._keyfield import CoordKeyField
from ._segy_text import (
    create_default_texthead,
    put_segy_texthead,
    _clean_texthead,
    trace_header_map_to_text,
)
from ._segy_globals import _ISEGY_MEASUREMENT_SYSTEM

TQDM_ARGS = dict(unit=" traces", desc="Writing to SEG-Y")


def _chunk_iterator(ds):
    # get the chunk selection labels
    if ds.chunks:
        ranges = {
            dim: tuple(split_into(ds[dim].values, ds.chunks[dim])) for dim in ds.chunks
        }

    else:
        ranges = {dim: [tuple(ds[dim].values)] for dim in ds.dims}
    lens = {key: len(val) for key, val in ranges.items()}
    order = tuple(ranges.keys())
    ranges = tuple(ranges.values())
    for selection in product(*map(range, tuple(lens.values()))):
        yield {order[i]: list(ranges[i][n]) for i, n in enumerate(selection)}


def _segy_freewriter(
    seisnc,
    segyfile,
    data_array="data",
    trace_header_map=None,
    silent=False,
    use_text=False,
    dead_trace_key=None,
    vert_dimension="twt",
    **dim_kwargs,
):
    """"""
    if trace_header_map:
        byte_fields = list(trace_header_map.values())
    else:
        byte_fields = []
    byte_fields += list(dim_kwargs.values())
    check_tracefield(byte_fields)

    try:
        coord_scalar = seisnc.coord_scalar
    except AttributeError:
        coord_scalar = 0
        seisnc.attrs["coord_scalar"] = coord_scalar
    coord_scalar_mult = np.power(abs(coord_scalar), np.sign(coord_scalar) * -1)
    seisnc.attrs["coord_scalar_mult"] = coord_scalar_mult

    # create empty trace header map if necessary
    if trace_header_map is None:
        trace_header_map = dict()
    else:  # check that values of thm in ds
        for key in trace_header_map:
            try:
                _ = seisnc[key]
            except KeyError:
                raise ValueError("keys of trace_header_map must be in seisnc")

    # remove duplicate dim_kwargs from trace_header_map
    for key in dim_kwargs:
        trace_header_map[key] = dim_kwargs[key]
    thm_vars = tuple(trace_header_map.keys())
    if dead_trace_key:
        thm_vars += (dead_trace_key,)

    dim_order = tuple(dim_kwargs.keys())

    # transfrom text if requested
    if use_text:
        text = {i + 1: line for i, line in enumerate(seisnc.text.split("\n"))}
        text = _clean_texthead(text, 80)
    else:
        text = None

    # create trace headers
    trace_layout = seisnc[data_array].isel(**{vert_dimension: 0})
    trace_headers = pd.DataFrame()
    for var in thm_vars:
        trace_headers[var] = (
            seisnc[var]
            .broadcast_like(trace_layout)
            .transpose(*dim_order, transpose_coords=True)
            .values.ravel()
        )

    # scale coords
    if CoordKeyField.cdp_x in trace_header_map:
        trace_headers[CoordKeyField.cdp_x] = (
            trace_headers[CoordKeyField.cdp_x] * seisnc.coord_scalar_mult
        )
    if CoordKeyField.cdp_y in trace_header_map:
        trace_headers[CoordKeyField.cdp_y] = (
            trace_headers[CoordKeyField.cdp_y] * seisnc.coord_scalar_mult
        )

    trace_headers = trace_headers.astype(np.int32)
    z0 = int(seisnc[vert_dimension].values[0])
    ntraces = trace_headers.shape[0]
    ns = seisnc[vert_dimension].values.size
    msys = _ISEGY_MEASUREMENT_SYSTEM[seisnc.seis.get_measurement_system()]
    spec = segyio.spec()
    spec.format = 1
    spec.iline = 181
    spec.xline = 185
    spec.samples = range(ns)

    if dead_trace_key:
        dead = ~trace_headers[dead_trace_key].astype(bool)
        spec.tracecount = dead.sum()
        live_index = trace_headers.where(dead).dropna().index
        tracen = pd.Series(
            index=live_index, data=np.arange(live_index.size), name="tracen"
        )
    else:
        spec.tracecount = ntraces
        tracen = pd.Series(
            index=trace_headers.index,
            name="tracen",
            data=np.arange(trace_headers.index.size),
        )

    # build trace headers
    trace_headers = trace_headers.astype(np.int32)
    trace_headers = trace_headers.join(
        tracen.astype(np.float32),
    )

    seisnc["tracen"] = trace_headers.set_index(list(dim_order))["tracen"].to_xarray()
    trace_headers = trace_headers.dropna(subset=["tracen"])
    trace_headers["tracen"] = trace_headers["tracen"].astype(np.int32)
    trace_headers = trace_headers.set_index("tracen")
    trace_headers = trace_headers[list(trace_header_map.keys())].rename(
        columns=trace_header_map
    )

    # header constants
    for byte, val in zip(
        [segyio.su.offset, segyio.su.ns, segyio.su.delrt, segyio.su.scalco],
        [1, ns, z0, int(seisnc.coord_scalar)],
    ):
        if byte not in trace_headers.columns:  # don't override user values
            trace_headers[byte] = val

    # write out to segy
    with segyio.create(segyfile, spec) as segyf:
        pbar = tqdm(total=spec.tracecount, disable=silent)  # , **TQDM_ARGS )
        for chk in _chunk_iterator(seisnc):
            _ = chk.pop(vert_dimension)
            # set order
            data = seisnc.sel(**chk).transpose(*dim_order, vert_dimension)
            # create missing trace mask and slicing tools
            mask = ~data.tracen.isnull().values.ravel()
            trace_locs = data.tracen.values.ravel()[mask].astype(int)
            trace_locs = tuple(
                map(
                    lambda x: slice(x[0], x[-1] + 1),
                    split_when(trace_locs, lambda x, y: y != x + 1),
                )
            )
            if data[data_array].size > 0:
                npdata = (
                    data[data_array].values.reshape(-1, ns)[mask, :].astype(np.float32)
                )
                n1 = 0
                for slc_set in trace_locs:
                    segyf.header[slc_set] = trace_headers[slc_set].to_dict(
                        orient="records"
                    )
                    n2 = n1 + slc_set.stop - slc_set.start
                    segyf.trace[slc_set] = npdata[n1:n2]
                    n1 = n2
                    pbar.update(slc_set.stop - slc_set.start)
        pbar.close()

        segyf.bin.update(
            # tsort=segyio.TraceSortingFormat.INLINE_SORTING,
            hdt=int(seisnc.sample_rate * 1000),
            hns=ns,
            mfeet=msys,
            jobid=1,
            lino=1,
            reno=1,
            ntrpr=ntraces,
            nart=ntraces,
            fold=1,
        )

    if not text:
        header_locs_text = trace_header_map_to_text(trace_header_map)

        # create text header
        overrides = {
            #     123456789012345678901234567890123456789012345678901234567890123456
            7: f"Original File: {seisnc.source_file}",
        }

        l1 = 40 - len(header_locs_text)
        overrides.update({l1 + i: line for i, line in enumerate(header_locs_text)})

        put_segy_texthead(
            segyfile,
            create_default_texthead(overrides),
        )
    else:
        put_segy_texthead(segyfile, text, line_counter=False)


def segy_freewriter(
    seisnc,
    segyfile,
    data_array="data",
    trace_header_map=None,
    use_text=False,
    dead_trace_key=None,
    vert_dimension="twt",
    chunk_spec=None,
    silent=False,
    **dim_kwargs,
):
    """Convert siesnc format (NetCDF4) to SEG-Y.

    Args:
        seisnc (xarray.Dataset, string): The input SEISNC file either a path or the in memory xarray.Dataset
        segyfile (string): The output SEG-Y file
        data_array (string): The Dataset variable name for the output volume.
        trace_header_map (dict, optional): Defaults to None. A dictionary of seisnc variables
            and byte locations. The variable will be written to the trace headers in the
            assigned byte location. By default CMP=23, cdp_x=181, cdp_y=185, iline=189,
            xline=193.
        use_text (book, optional): Use the seisnc text for the EBCIDC output. This text usally comes from
            the loaded SEG-Y file and may not match the segysak SEG-Y output. Defaults to False and writes
            the default segysak EBCIDC
        dead_trace_key (str): The key for the Dataset variable to use for trace output filter.
        vert_dimension (str): Data dimension to output, defaults to 'twt' or 'depth' whichever
            is present.
        chunk_spec (dict, optional): Xarray open_dataset chunking spec.
        silent (bool, optional): Turn off progress reporting. Defaults to False.
        dim_kwargs (int): The dimension/byte location pairs to output dimensions to. The number of dim_kwargs should be
            equal to the number of dimensions on the output data_array. The sort order will be as per the order passed
            to the function.
    """
    kwargs = dict(
        data_array=data_array,
        trace_header_map=trace_header_map,
        silent=silent,
        use_text=use_text,
        dead_trace_key=dead_trace_key,
        vert_dimension=vert_dimension,
    )

    if isinstance(seisnc, xr.Dataset):
        _segy_freewriter(
            seisnc,
            segyfile,
            **kwargs,
            **dim_kwargs,
        )
    else:
        ncfile = seisnc
        with open_seisnc(ncfile, chunks=chunk_spec) as seisnc:
            _segy_freewriter(
                seisnc,
                segyfile,
                **kwargs,
                **dim_kwargs,
            )



from collections import defaultdict

_SEGY_MEASUREMENT_SYSTEM = defaultdict(lambda: 0)
_SEGY_MEASUREMENT_SYSTEM[1] = "m"
_SEGY_MEASUREMENT_SYSTEM[2] = "ft"
_ISEGY_MEASUREMENT_SYSTEM = defaultdict(lambda: 0, m=1, ft=2)



# pylint: disable=invalid-name, unused-variable
"""Functions to interact with segy data
"""

import pathlib
from numpy.core.fromnumeric import trace

import segyio

# import netCDF
import h5netcdf
from attrdict import AttrDict
import numpy as np
import xarray as xr

from ._segy_core import tqdm, check_tracefield

TQDM_ARGS = dict(unit_scale=True, unit=" traces")
PERCENTILES = [0, 0.1, 10, 50, 90, 99.9, 100]

KNOWN_BYTES = dict(
    standard_3d=dict(iline=181, xline=185, cdpx=189, cdpy=193),
    standard_3d_gath=dict(iline=181, xline=185, cdpx=189, cdpy=193, offset=37),
    standard_2d=dict(cdp=21, cdpx=189, cdpy=193),
    standard_2d_gath=dict(cdp=21, cdpx=189, cdpy=193, offset=37),
    petrel_3d=dict(iline=5, xline=21, cdpx=73, cdpy=77),
    petrel_2d=dict(cdp=21, cdpx=73, cdpy=77),
)

from segysak._keyfield import (
    CoordKeyField,
    AttrKeyField,
    VariableKeyField,
    VerticalKeyDim,
    DimensionKeyField,
)
from segysak._seismic_dataset import (
    create_seismic_dataset,
    _dataset_coordinate_helper,
)

# from segysak.seisnc import create_empty_seisnc, set_seisnc_dims
from segysak.tools import check_crop, check_zcrop
from segysak._accessor import open_seisnc
from segysak._core import FrozenDict

KNOWN_BYTES = FrozenDict(KNOWN_BYTES)

from ._segy_headers import (
    segy_bin_scrape,
    segy_header_scrape,
    what_geometry_am_i,
    _active_tracefield_segyio,
)
from ._segy_text import get_segy_texthead
from ._segy_globals import _SEGY_MEASUREMENT_SYSTEM


class SegyLoadError(Exception):
    def __init__(self, message, errors):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors


def _get_tf(var):
    return str(segyio.TraceField(var))


def _header_to_index_mapping(series):
    as_unique = series.unique()
    sorted_unique = np.sort(as_unique)
    mapper = {val: i for i, val in enumerate(sorted_unique)}
    return series.map(mapper.get)


def _segy3d_ncdf(
    segyfile,
    ncfile,
    segyio_kwargs,
    n0,
    ns,
    head_df,
    head_loc,
    vert_domain="TWT",
    silent=False,
):
    """Helper function to load 3d data into a netcdf4 file. This function should
    stream data to disk avoiding excessive memory usage.
    """

    if vert_domain == "TWT":
        dims = DimensionKeyField.threed_twt
    elif vert_domain == "DEPTH":
        dims = DimensionKeyField.threed_depth
    else:
        raise ValueError(f"Unknown vert_domain: {vert_domain}")

    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf, h5netcdf.File(
        ncfile, "a"
    ) as seisnc:

        segyf.mmap()

        # create data variable
        seisnc_data = seisnc.create_variable(VariableKeyField.data, dims, np.float32)

        seisnc.flush()

        # work out fast and slow dir
        if head_df[head_loc.xline].diff().min() < 0:
            contig_dir = head_loc.iline
            broken_dir = head_loc.xline
            # slicer = lambda x, y: slice(x, y, ...)
        else:
            contig_dir = head_loc.xline
            broken_dir = head_loc.iline
            # slicer = lambda x, y: slice(y, x, ...)

        print(f"Fast direction is {broken_dir}")

        pb = tqdm(
            total=segyf.tracecount, desc="Converting SEGY", disable=silent, **TQDM_ARGS
        )

        percentiles = np.zeros_like(PERCENTILES)
        for _, grp in head_df.groupby(contig_dir):
            for trc, val in grp.iterrows():
                i1, i2 = val[["il_index", "xl_index"]].values.astype(int)
                seisnc_data[i1, i2, :] = segyf.trace[trc][n0 : ns + 1].astype(
                    np.float32
                )
                percentiles = (
                    np.percentile(segyf.trace[trc][n0 : ns + 1], PERCENTILES)
                    + percentiles
                ) / 2.0

                pb.update()
        pb.close()
        seisnc.attrs[AttrKeyField.percentiles] = list(percentiles)
        seisnc.attrs[AttrKeyField.coord_scalar] = head_df.SourceGroupScalar.median()
        seisnc.flush()

    return ncfile


def _segy3dps_ncdf(
    segyfile,
    ncfile,
    segyio_kwargs,
    n0,
    ns,
    head_df,
    head_loc,
    vert_domain="TWT",
    silent=False,
):
    """Helper function to load 3d pre-stack data into a netcdf4 file. This function should
    stream data to disk avoiding excessive memory usage.
    """

    if vert_domain == "TWT":
        dims = DimensionKeyField.threed_ps_twt
    elif vert_domain == "DEPTH":
        dims = DimensionKeyField.threed_ps_depth
    else:
        raise ValueError(f"Unknown vert_domain: {vert_domain}")

    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf, h5netcdf.File(
        ncfile, "a"
    ) as seisnc:

        segyf.mmap()

        # create data variable
        seisnc_data = seisnc.create_variable(VariableKeyField.data, dims, np.float32)

        seisnc.flush()

        # work out fast and slow dir
        if head_df[head_loc.xline].diff().min() < 0:
            contig_dir = head_loc.iline
            broken_dir = head_loc.xline
            slicer = lambda x, y: slice(x, y, ...)
        else:
            contig_dir = head_loc.xline
            broken_dir = head_loc.iline
            slicer = lambda x, y: slice(y, x, ...)

        print(f"Fast direction is {broken_dir}")

        pb = tqdm(
            total=segyf.tracecount, desc="Converting SEGY", disable=silent, **TQDM_ARGS
        )

        percentiles = np.zeros_like(PERCENTILES)

        for contig, grp in head_df.groupby(contig_dir):
            for trc, val in grp.iterrows():
                seisnc_data[val.il_index, val.xl_index, :, val.off_index] = segyf.trace[
                    trc
                ][n0 : ns + 1]

                percentiles = (
                    np.percentile(segyf.trace[trc][n0 : ns + 1], PERCENTILES)
                    + percentiles
                ) / 2.0

                pb.update()
        pb.close()

        seisnc.attrs[AttrKeyField.percentiles] = list(percentiles)
        seisnc.flush()

    return ncfile


def _segy3d_xr(
    segyfile,
    ds,
    segyio_kwargs,
    n0,
    ns,
    head_df,
    head_loc,
    vert_domain="TWT",
    silent=False,
):
    """Helper function to load 3d data into an xarray with the seisnc form."""

    if vert_domain == "TWT":
        dims = DimensionKeyField.threed_twt
    elif vert_domain == "DEPTH":
        dims = DimensionKeyField.threed_depth
    else:
        raise ValueError(f"Unknown vert_domain: {vert_domain}")

    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf:

        segyf.mmap()

        # work out fast and slow dir
        if head_df[head_loc.xline].diff().min() < 0:
            contig_dir = head_loc.iline
            broken_dir = head_loc.xline
            # slicer = lambda x, y: slice(x, y, ...)
        else:
            contig_dir = head_loc.xline
            broken_dir = head_loc.iline
            # slicer = lambda x, y: slice(y, x, ...)

        print(f"Fast direction is {broken_dir}")

        pb = tqdm(
            total=segyf.tracecount, desc="Converting SEGY", disable=silent, **TQDM_ARGS
        )
        shape = [ds.dims[d] for d in dims]
        volume = np.zeros(shape, dtype=np.float32)
        percentiles = np.zeros_like(PERCENTILES)

        for contig, grp in head_df.groupby(contig_dir):
            for trc, val in grp.iterrows():
                volume[int(val.il_index), int(val.xl_index), :] = segyf.trace[trc][
                    n0 : ns + 1
                ]
                percentiles = (
                    np.percentile(segyf.trace[trc][n0 : ns + 1], PERCENTILES)
                    + percentiles
                ) / 2.0
                pb.update()
        pb.close()

    ds[VariableKeyField.data] = (dims, volume)
    ds.attrs[AttrKeyField.percentiles] = list(percentiles)

    return ds


def _segy3dps_xr(
    segyfile,
    ds,
    segyio_kwargs,
    n0,
    ns,
    head_df,
    head_loc,
    vert_domain="TWT",
    silent=False,
):
    """Helper function to load 3d pre-stack data into an xarray with the seisnc form."""

    if vert_domain == "TWT":
        dims = DimensionKeyField.threed_ps_twt
    elif vert_domain == "DEPTH":
        dims = DimensionKeyField.threed_ps_depth
    else:
        raise ValueError(f"Unknown vert_domain: {vert_domain}")

    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf:

        segyf.mmap()

        # work out fast and slow dir
        if head_df[head_loc.xline].diff().min() < 0:
            contig_dir = head_loc.iline
            broken_dir = head_loc.xline
            slicer = lambda x, y: slice(x, y, ...)
        else:
            contig_dir = head_loc.xline
            broken_dir = head_loc.iline
            slicer = lambda x, y: slice(y, x, ...)

        print(f"Fast direction is {broken_dir}")

        pb = tqdm(
            total=segyf.tracecount, desc="Converting SEGY", disable=silent, **TQDM_ARGS
        )
        shape = [ds.dims[d] for d in dims]
        volume = np.zeros(shape, dtype=np.float32)
        percentiles = np.zeros_like(PERCENTILES)

        for contig, grp in head_df.groupby(contig_dir):
            for trc, val in grp.iterrows():
                volume[
                    int(val.il_index), int(val.xl_index), :, int(val.off_index)
                ] = segyf.trace[trc][n0 : ns + 1]
                percentiles = (
                    np.percentile(segyf.trace[trc][n0 : ns + 1], PERCENTILES)
                    + percentiles
                ) / 2.0
                pb.update()
        pb.close()

    ds[VariableKeyField.data] = (dims, volume)
    ds.attrs[AttrKeyField.percentiles] = list(percentiles)

    return ds


def _3dsegy_loader(
    segyfile,
    head_df,
    head_bin,
    head_loc,
    ncfile=None,
    vert_domain="TWT",
    data_type="AMP",
    crop=None,
    zcrop=None,
    silent=False,
    return_geometry=False,
    **segyio_kwargs,
):
    """Convert SEG-Y data to Xarray or Netcdf4

    This is a helper function for segy_loader. Users should use that function
    directly to load all segy data.

    """

    # get vertical sample ranges
    n0 = 0
    nsamp = head_bin["Samples"]
    ns0 = head_df.DelayRecordingTime.min()

    # short way to get inlines/xlines
    ilines = head_df[head_loc.iline].unique()
    xlines = head_df[head_loc.xline].unique()
    ilines = np.sort(ilines)
    xlines = np.sort(xlines)
    iline_index_map = {il: i for i, il in enumerate(ilines)}
    xline_index_map = {xl: i for i, xl in enumerate(xlines)}
    head_df.loc[:, "il_index"] = head_df[head_loc.iline].map(iline_index_map.get)
    head_df.loc[:, "xl_index"] = head_df[head_loc.xline].map(xline_index_map.get)

    # binary header translation
    ns = head_bin["Samples"]
    sample_rate = head_bin["Interval"] / 1000.0
    msys = _SEGY_MEASUREMENT_SYSTEM[head_bin["MeasurementSystem"]]

    # for offset
    if head_loc.offset is not None:
        offsets = head_df[head_loc.offset].unique()
        offsets = np.sort(offsets)
        offset_index_map = {off: i for i, off in enumerate(offsets)}
        head_df["off_index"] = head_df[head_loc.offset].map(offset_index_map.get)
    else:
        offsets = None

    if zcrop is not None:
        zcrop = check_zcrop(zcrop, [0, ns])
        n0, ns = zcrop
        ns0 = sample_rate * n0
        nsamp = ns - n0 + 1
    vert_samples = np.arange(ns0, ns0 + sample_rate * nsamp, sample_rate)

    builder, domain = _dataset_coordinate_helper(
        vert_samples, vert_domain, iline=ilines, xline=xlines, offset=offsets
    )

    ds = create_seismic_dataset(**builder)

    # create_seismic_dataset(d1=ni, d2=nx, d3=nsamp)
    text = get_segy_texthead(segyfile, **segyio_kwargs)
    ds.attrs[AttrKeyField.text] = text
    ds.attrs[AttrKeyField.source_file] = pathlib.Path(segyfile).name
    ds.attrs[AttrKeyField.measurement_system] = msys
    ds.attrs[AttrKeyField.sample_rate] = sample_rate
    ds.attrs[AttrKeyField.coord_scalar] = head_df.SourceGroupScalar.median()

    if ncfile is not None and return_geometry == False:
        ds.seisio.to_netcdf(ncfile)
    elif return_geometry:
        # return geometry -> e.g. don't process segy traces
        return ds
    else:
        ncfile = ds

    segyio_kwargs.update(
        dict(ignore_geometry=True, iline=head_loc.iline, xline=head_loc.xline)
    )

    # not prestack data
    if head_loc.offset is None and not isinstance(ncfile, xr.Dataset):
        ds = _segy3d_ncdf(
            segyfile,
            ncfile,
            segyio_kwargs,
            n0,
            ns,
            head_df,
            head_loc,
            vert_domain=vert_domain,
            silent=silent,
        )

    # not prestack data load into memory
    if head_loc.offset is None and isinstance(ncfile, xr.Dataset):
        ds = _segy3d_xr(
            segyfile,
            ncfile,
            segyio_kwargs,
            n0,
            ns,
            head_df,
            head_loc,
            vert_domain=vert_domain,
            silent=silent,
        )

    # prestack data
    if head_loc.offset is not None and not isinstance(ncfile, xr.Dataset):
        ds = _segy3dps_ncdf(
            segyfile,
            ncfile,
            segyio_kwargs,
            n0,
            ns,
            head_df,
            head_loc,
            vert_domain=vert_domain,
            silent=silent,
        )

    # prestack data load into memory
    if head_loc.offset is not None and isinstance(ncfile, xr.Dataset):
        ds = _segy3dps_xr(
            segyfile,
            ncfile,
            segyio_kwargs,
            n0,
            ns,
            head_df,
            head_loc,
            vert_domain=vert_domain,
            silent=silent,
        )

    return ds


def _segy2d_xr(
    segyfile,
    ds,
    segyio_kwargs,
    n0,
    ns,
    head_df,
    head_loc,
    vert_domain="TWT",
    silent=False,
):
    """Helper function to load 2d data into an xarray with the seisnc form."""

    if vert_domain == "TWT":
        dims = DimensionKeyField.twod_twt
    elif vert_domain == "DEPTH":
        dims = DimensionKeyField.twod_depth
    else:
        raise ValueError(f"Unknown vert_domain: {vert_domain}")

    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf:

        segyf.mmap()

        pb = tqdm(
            total=segyf.tracecount, desc="Converting SEGY", disable=silent, **TQDM_ARGS
        )
        shape = [ds.dims[d] for d in dims]
        volume = np.zeros(shape, dtype=np.float32)
        percentiles = np.zeros_like(PERCENTILES)

        # this can probably be done as a block - leaving for now just incase sorting becomes an issue
        for trc, val in head_df.iterrows():
            volume[int(val.cdp_index), :] = segyf.trace[trc][n0 : ns + 1]
            pb.update()
            percentiles = (
                np.percentile(segyf.trace[trc][n0 : ns + 1], PERCENTILES) + percentiles
            ) / 2.0
        pb.close()

    ds[VariableKeyField.data] = (
        dims,
        volume[:, n0 : ns + 1],
    )
    ds.attrs[AttrKeyField.percentiles] = list(percentiles)

    return ds


def _segy2d_ps_xr(
    segyfile,
    ds,
    segyio_kwargs,
    n0,
    ns,
    head_df,
    head_loc,
    vert_domain="TWT",
    silent=False,
):
    """Helper function to load 2d data into an xarray with the seisnc form."""

    if vert_domain == "TWT":
        dims = DimensionKeyField.twod_ps_twt
    elif vert_domain == "DEPTH":
        dims = DimensionKeyField.twod_ps_depth
    else:
        raise ValueError(f"Unknown vert_domain: {vert_domain}")

    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf:

        segyf.mmap()

        pb = tqdm(
            total=segyf.tracecount, desc="Converting SEGY", disable=silent, **TQDM_ARGS
        )
        shape = [ds.dims[d] for d in dims]
        volume = np.zeros(shape, dtype=np.float32)
        percentiles = np.zeros_like(PERCENTILES)

        # this can probably be done as a block - leaving for now just incase sorting becomes an issue
        for trc, val in head_df.iterrows():
            volume[int(val.cdp_index), :, int(val.off_index)] = segyf.trace[trc][
                n0 : ns + 1
            ]
            percentiles = (
                np.percentile(segyf.trace[trc][n0 : ns + 1], PERCENTILES) + percentiles
            ) / 2.0
            pb.update()
        pb.close()

    ds[VariableKeyField.data] = (
        dims,
        volume,
    )
    ds.attrs[AttrKeyField.percentiles] = list(percentiles)

    return ds


def _2dsegy_loader(
    segyfile,
    head_df,
    head_bin,
    head_loc,
    ncfile=None,
    vert_domain="TWT",
    data_type="AMP",
    crop=None,
    zcrop=None,
    silent=False,
    return_geometry=False,
    **segyio_kwargs,
):
    """Convert SEG-Y data to Xarray or Netcdf4

    This is a helper function for segy_loader. Users should use that function
    directly to load all segy data.

    """

    # get names of columns where stuff we want is
    if head_loc.cdp is None:
        cdp = 21
        head_loc.cdp = _get_tf(cdp)
        head_df[head_loc.cdp] = head_df.index.values

    # get vertical sample ranges
    n0 = 0
    nsamp = head_bin["Samples"]
    ns0 = head_df.DelayRecordingTime.min()

    # short way to get cdps
    cdps = head_df[head_loc.cdp].unique()
    cdps = np.sort(cdps)
    head_df["cdp_index"] = _header_to_index_mapping(head_df[head_loc.cdp])

    # binary header translation
    nsamp = head_bin["Samples"]
    sample_rate = head_bin["Interval"] / 1000.0
    msys = _SEGY_MEASUREMENT_SYSTEM[head_bin["MeasurementSystem"]]

    # for offset
    if head_loc.offset is not None:
        offsets = head_df[head_loc.offset].unique()
        offsets = np.sort(offsets)
        head_df["off_index"] = _header_to_index_mapping(head_df[head_loc.offset])
    else:
        offsets = None

    if zcrop is not None:
        zcrop = check_zcrop(zcrop, [0, nsamp])
        n0, nsamp = zcrop
        ns0 = sample_rate * n0
        nsamp = nsamp - n0 + 1
    vert_samples = np.arange(ns0, ns0 + sample_rate * nsamp, sample_rate)

    builder, domain = _dataset_coordinate_helper(
        vert_samples, vert_domain, cdp=cdps, offset=offsets
    )

    ds = create_seismic_dataset(**builder)

    # create_seismic_dataset(d1=ni, d2=nx, d3=nsamp)
    text = get_segy_texthead(segyfile, **segyio_kwargs)
    ds.attrs[AttrKeyField.text] = text
    ds.attrs[AttrKeyField.source_file] = pathlib.Path(segyfile).name
    ds.attrs[AttrKeyField.measurement_system] = msys
    ds.attrs[AttrKeyField.sample_rate] = sample_rate
    ds.attrs[AttrKeyField.coord_scalar] = head_df.SourceGroupScalar.median()

    if ncfile is not None and return_geometry == True:
        ds.seisio.to_netcdf(ncfile)
        return ds
    elif return_geometry:
        return ds

    segyio_kwargs.update(dict(ignore_geometry=True))

    # stacked data
    if head_loc.offset is None:
        ds = _segy2d_xr(
            segyfile,
            ds,
            segyio_kwargs,
            n0,
            nsamp,
            head_df,
            head_loc,
            vert_domain=vert_domain,
            silent=silent,
        )

    # # prestack data
    if head_loc.offset is not None:
        ds = _segy2d_ps_xr(
            segyfile,
            ds,
            segyio_kwargs,
            n0,
            nsamp,
            head_df,
            head_loc,
            vert_domain=vert_domain,
            silent=silent,
        )

    if ncfile is not None:
        ds.seisio.to_netcdf(ncfile)
        del ds
        ds = ncfile

    return ds


def well_known_byte_locs(name):
    """Return common bytes position kwargs_dict for segy_loader and segy_converter.

    Returns a dict containing the byte locations for well known SEG-Y variants in the wild.

    Args:
        name (str): Takes one of keys from KNOWN_BYTES

    Returns:
        dict: A dictionary of SEG-Y byte positions.

    Example:

    Use the output of this function to unpack arguments into ``segy_loader``

    >>> seismic = segy_loader(filepath, **well_known_byte_locs('petrel_3d'))

    """
    try:
        return KNOWN_BYTES[name]
    except KeyError:
        raise ValueError(
            f"No byte locatons for {name}, select from {list(KNOWN_BYTES.keys())}"
        )


def _loader_converter_checks(cdp, iline, xline, extra_byte_fields):
    # Input sanity checks
    if cdp is not None and (iline is not None or xline is not None):
        raise ValueError("cdp cannot be defined with iline and xiline")

    if iline is None and xline is not None:
        raise ValueError("iline must be defined with xline")

    if xline is None and iline is not None:
        raise ValueError("xline must be defined with iline")

    if isinstance(extra_byte_fields, list):
        extra_byte_fields = {
            _get_tf(field): _get_tf(field) for field in extra_byte_fields
        }
    elif isinstance(extra_byte_fields, dict):
        extra_byte_fields = {
            key: _get_tf(field) for key, field in extra_byte_fields.items()
        }
    elif extra_byte_fields is None:
        extra_byte_fields = dict()
    else:
        raise ValueError("Unknown type for extra_byte_fields")
    return extra_byte_fields


def _loader_converter_header_handling(
    segyfile,
    cdp=None,
    iline=None,
    xline=None,
    cdpx=None,
    cdpy=None,
    offset=None,
    vert_domain="TWT",
    data_type="AMP",
    ix_crop=None,
    cdp_crop=None,
    xy_crop=None,
    z_crop=None,
    return_geometry=False,
    silent=False,
    extra_byte_fields=None,
    head_df=None,
    optimised_load=False,
    **segyio_kwargs,
):

    if optimised_load:
        scrape_bytes = [
            val for val in [cdp, iline, xline, cdpx, cdpy, offset] if val is not None
        ]
        scrape_bytes += list(extra_byte_fields.values())
        scrape_bytes += [
            _active_tracefield_segyio()["DelayRecordingTime"],
            _active_tracefield_segyio()["SourceGroupScalar"],
        ]
    else:
        scrape_bytes = None

    if head_df is None:
        # Start by scraping the headers.
        head_df = segy_header_scrape(
            segyfile, silent=silent, bytes_filter=scrape_bytes, **segyio_kwargs
        )

    head_bin = segy_bin_scrape(segyfile, **segyio_kwargs)
    head_loc = AttrDict(
        dict(cdp=cdp, offset=offset, iline=iline, xline=xline, cdpx=cdpx, cdpy=cdpy)
    )

    if all(map(lambda x: x is None, (cdp, iline, xline, offset))):
        # lets try and guess the data types if no hints given
        try:
            head_loc.update(what_geometry_am_i(head_df))
        except (KeyError, ValueError):
            print("Couldn't determine geometry, will load traces as flat 2D.")
            pass

    head_loc = AttrDict(
        {
            key: (_get_tf(val) if val is not None else None)
            for key, val in head_loc.items()
        }
    )

    if all(v is not None for v in (head_loc.cdpx, head_loc.cdpy)):
        extra_byte_fields[CoordKeyField.cdp_x] = head_loc.cdpx
        extra_byte_fields[CoordKeyField.cdp_y] = head_loc.cdpy

        # Scale Coordinates
        coord_scalar = head_df.SourceGroupScalar.median()
        coord_scalar_mult = np.power(abs(coord_scalar), np.sign(coord_scalar))
        head_df[head_loc.cdpx] = head_df[head_loc.cdpx].astype(np.float32)
        head_df[head_loc.cdpy] = head_df[head_loc.cdpy].astype(np.float32)
        head_df[head_loc.cdpx] = head_df[head_loc.cdpx] * coord_scalar_mult * 1.0
        head_df[head_loc.cdpy] = head_df[head_loc.cdpy] * coord_scalar_mult * 1.0

    # TODO: might need to scale offsets as well? This isn't available in segy standard

    # Cropping
    if cdp_crop and cdp is not None:  # 2d cdp cropping
        crop_min, crop_max = check_crop(
            cdp_crop, [head_df[head_loc.cdp].min(), head_df[head_loc.cdp].max()]
        )

        head_df = head_df.query(
            "@head_loc.cdp >= @crop_min & @head_loc.cdp <= @crop_max"
        )

    if ix_crop is not None and cdp is None:  # 3d inline/xline cropping
        il_min, il_max, xl_min, xl_max = check_crop(
            ix_crop,
            [
                head_df[head_loc.iline].min(),
                head_df[head_loc.iline].max(),
                head_df[head_loc.xline].min(),
                head_df[head_loc.xline].max(),
            ],
        )
        query = " & ".join(
            [
                f"{head_loc.iline} >= @il_min",
                f"{head_loc.iline} <= @il_max",
                f"{head_loc.xline} >= @xl_min",
                f"{head_loc.xline} <= @xl_max",
            ]
        )
        head_df = head_df.query(query).copy(deep=True)

    # TODO: -> Could implement some cropping with a polygon here
    if xy_crop is not None and cdp is None:
        x_min, x_max, y_min, y_max = check_crop(
            xy_crop,
            [
                head_df[head_loc.cdpx].min(),
                head_df[head_loc.cdpx].max(),
                head_df[head_loc.cdpy].min(),
                head_df[head_loc.cdpy].max(),
            ],
        )
        query = " & ".join(
            [
                f"{head_loc.cdpx} >= @x_min",
                f"{head_loc.cdpx} <= @x_max",
                f"{head_loc.cdpy} >= @y_min",
                f"{head_loc.cdpy} <= @y_max",
            ]
        )
        head_df = head_df.query(query).copy(deep=True)

    return head_df, head_bin, head_loc


def _loader_converter_write_headers(
    ds, head_df, indexer, dims, extra_byte_fields, is3d2d=True
):

    if not isinstance(ds, xr.Dataset):
        ds = open_seisnc(ds)

    # we have some some geometry to assign headers to
    if is3d2d:
        head_ds = head_df.set_index(indexer).to_xarray()
        for key, field in extra_byte_fields.items():
            ds[key] = (dims, head_ds[field].values)
    # geometry is not known
    else:
        for key, field in extra_byte_fields.items():
            ds[key] = (dims, head_df[field].values)

    coords = (CoordKeyField.cdp_x, CoordKeyField.cdp_y)
    if set(coords).issubset(ds.variables.keys()):
        ds = ds.set_coords(coords)

    return ds


def segy_freeloader(
    segyfile,
    vert_domain="TWT",
    data_type="AMP",
    return_geometry=False,
    silent=False,
    extra_byte_fields=None,
    head_df=None,
    segyio_kwargs=None,
    **dim_kwargs,
):
    """Freeform loader for SEG-Y data. This loader allows you to load SEG-Y into
    an xarray.Dataset using an arbitrary number of header locations to create
    othogonal dimensions. This is an eager loader and will transfer the entire
    SEG-Y and requested header information to memory.

    From the dimension header locations specified the freeloader will try to
    create a Dataset where each trace is assigned to a dimension.

    Args:
        segyfile (string): The SEG-Y file/path.
        vert_domain (str, optional): One of ('TWT', 'DEPTH'). Defaults to 'TWT'.
        data_type (str, optional): Defaults to "AMP".
        return_geometry (bool, optional): If true, just returned the empty
            dataset based upon the calcuated header geometry. Defaults to False.
        silent (bool, optional): Turn off progress bars. Defaults to False.
        extra_byte_fields (dict, optional): Additional header information to
            load into the Dataset. Defaults to None.
        head_df (pandas.DataFrame): The DataFrame output from `segy_header_scrape`.
            This DataFrame can be filtered by the user
            to load select trace sets. Trace loading is based upon the DataFrame index.
        segyio_kwargs (dict, optional): Extra keyword arguments for segyio.open
        **dim_kwargs: Dimension names and byte location pairs.
    """
    if segyio_kwargs is None:
        segyio_kwargs = dict()

    if head_df is None:
        # Start by scraping the headers.
        head_df = segy_header_scrape(segyfile, silent=silent, **segyio_kwargs)

    head_bin = segy_bin_scrape(segyfile, **segyio_kwargs)

    # get vertical sample ranges
    n0 = 0
    nsamp = head_bin["Samples"]
    ns0 = head_df.DelayRecordingTime.min()

    # binary header translation
    nsamp = head_bin["Samples"]
    sample_rate = head_bin["Interval"] / 1000.0
    msys = _SEGY_MEASUREMENT_SYSTEM[head_bin["MeasurementSystem"]]
    vert_samples = np.arange(ns0, ns0 + sample_rate * nsamp, sample_rate, dtype=int)

    # creating dimensions and new dataset
    dims = dict()
    dim_index_names = list()
    dim_fields = list()
    for dim in dim_kwargs:
        trace_field = str(segyio.TraceField(dim_kwargs[dim]))
        if trace_field == "Unknown Enum":
            raise ValueError(f"{dim}:{dim_kwargs[dim]} was not a valid byte header")
        dim_fields.append(trace_field)
        as_unique = head_df[trace_field].unique()
        dims[dim] = np.sort(as_unique)
        d_map = {dval: i for i, dval in enumerate(as_unique)}
        index_name = f"{dim}_index"
        dim_index_names.append(index_name)
        head_df.loc[:, index_name] = head_df[trace_field].map(d_map)

    if (
        head_df[dim_index_names].shape
        != head_df[dim_index_names].drop_duplicates().shape
    ):
        raise ValueError(
            "The selected dimensions results in multiple traces per "
            "dimension location, add additional dimensions or use "
            "trace numbering to load as 2D."
        )

    builder, domain = _dataset_coordinate_helper(vert_samples, vert_domain, **dims)
    ds = create_seismic_dataset(**builder)

    # getting attributes
    text = get_segy_texthead(segyfile, **segyio_kwargs)
    ds.attrs[AttrKeyField.text] = text
    ds.attrs[AttrKeyField.source_file] = pathlib.Path(segyfile).name
    ds.attrs[AttrKeyField.measurement_system] = msys
    ds.attrs[AttrKeyField.sample_rate] = sample_rate

    # map extra byte fields into ds
    if extra_byte_fields is not None:
        to_add = list()
        for name, byte in extra_byte_fields.items():
            trace_field = str(segyio.TraceField(byte))
            if trace_field == "Unknown Enum":
                raise ValueError(f"{name}:{byte} was not a valid byte header")
            to_add.append(trace_field)
        to_add = to_add + dim_fields
        extras = head_df[to_add].set_index(dim_fields).to_xarray()
        extras = extras.rename_dims(
            {b: a for a, b in zip(dim_kwargs, dim_fields) if a != b}
        )
        for name, xtr in zip(extra_byte_fields, to_add):
            ds[name] = extras[xtr]

    if return_geometry:
        # return geometry -> e.g. don't process segy traces
        return ds

    segyio_kwargs.update(dict(ignore_geometry=True))

    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf:

        segyf.mmap()
        shape = [ds.dims[d] for d in dim_kwargs] + [vert_samples.size]
        volume = np.zeros(shape, dtype=np.float32)

        # this can probably be done as a block - leaving for now just incase sorting becomes an issue
        indexes = tuple([head_df[idx].values for idx in dim_index_names])
        volume[indexes] = segyf.trace.raw[:][head_df.index.values]

    percentiles = np.percentile(volume, PERCENTILES)
    ds[VariableKeyField.data] = (
        list(dim_kwargs) + [VerticalKeyDim[domain]],
        volume,
    )
    ds.attrs[AttrKeyField.percentiles] = list(percentiles)

    return ds


def segy_loader(
    segyfile,
    cdp=None,
    iline=None,
    xline=None,
    cdpx=None,
    cdpy=None,
    offset=None,
    vert_domain="TWT",
    data_type="AMP",
    ix_crop=None,
    cdp_crop=None,
    xy_crop=None,
    z_crop=None,
    return_geometry=False,
    silent=False,
    extra_byte_fields=None,
    head_df=None,
    **segyio_kwargs,
):
    """Load SEG-Y file into xarray.Dataset

    The output dataset has the following structure
        Dimensions:
            cdp/iline - CDP or Inline axis
            xline - Xline axis
            twt/depth - The vertical axis
            offset - Offset/Angle Axis
        Coordinates:
            iline - The inline numbering
            xline - The xline numbering
            cdp_x - Eastings
            cdp_y - Northings
            cdp - Trace Number for 2d
        Variables
            data - The data volume
        Attributes:
            ns - number of samples vertical
            sample_rate - sample rate in ms/m
            test - text header
            measurement_system : m/ft
            source_file : segy source
            srd : seismic reference datum
            percentiles : data amplitude percentiles
            coord_scalar : from trace headers

    Args:
        segyfile (str): Input segy file path
        cdp (int, optional): The CDP byte location, usually 21.
        iline (int, optional): Inline byte location, usually 189
        xline (int, optional): Cross-line byte location, usually 193
        cdpx (int, optional): UTMX byte location, usually 181
        cdpy (int, optional): UTMY byte location, usually 185
        offset (int, optional): Offset/angle byte location
        vert_domain (str, optional): Vertical sampling domain. One of ['TWT', 'DEPTH']. Defaults to 'TWT'.
        data_type (str, optional): Data type ['AMP', 'VEL']. Defaults to 'AMP'.
        ix_crop (list, optional): List of minimum and maximum inline and crossline to output.
            Has the form '[min_il, max_il, min_xl, max_xl]'. Ignored for 2D data.
        cdp_crop (list, optional): List of minimum and maximum cmp values to output.
            Has the form '[min_cmp, max_cmp]'. Ignored for 3D data.
        xy_crop (list, optional): List of minimum and maximum cdp_x and cdp_y to output.
            Has the form '[min_x, max_x, min_y, max_y]'. Ignored for 2D data.
        z_crop (list, optional): List of minimum and maximum vertical samples to output.
            Has the form '[min, max]'.
        return_geometry (bool, optional): If true returns an xarray.dataset which doesn't contain data but mirrors
            the input volume header information.
        silent (bool): Disable progress bar.
        extra_byte_fields (list/mapping): A list of int or mapping of byte fields that should be returned as variables in the dataset.
        head_df (pandas.DataFrame): The DataFrame output from `segy_header_scrape`. This DataFrame can be filtered by the user
            to load select trace sets. Trace loading is based upon the DataFrame index.
        **segyio_kwargs: Extra keyword arguments for segyio.open

    Returns:
        xarray.Dataset: If ncfile keyword is specified returns open handle to disk netcdf4,
            otherwise the data in memory. If return_geometry is True does not load trace data and
            returns headers in geometry.
    """
    extra_byte_fields = _loader_converter_checks(cdp, iline, xline, extra_byte_fields)

    head_df, head_bin, head_loc = _loader_converter_header_handling(
        segyfile,
        cdp=cdp,
        iline=iline,
        xline=xline,
        cdpx=cdpx,
        cdpy=cdpy,
        offset=offset,
        vert_domain=vert_domain,
        data_type=data_type,
        ix_crop=ix_crop,
        cdp_crop=cdp_crop,
        xy_crop=xy_crop,
        z_crop=z_crop,
        return_geometry=return_geometry,
        silent=silent,
        extra_byte_fields=extra_byte_fields,
        head_df=head_df,
        **segyio_kwargs,
    )

    byte_loc = [
        bytel for bytel in [cdp, iline, xline, cdpx, cdpy, offset] if bytel is not None
    ]
    check_tracefield(byte_loc)

    common_args = (segyfile, head_df, head_bin, head_loc)

    common_kwargs = dict(
        zcrop=z_crop,
        vert_domain=vert_domain,
        data_type=data_type,
        return_geometry=return_geometry,
        silent=silent,
    )

    # 3d data needs iline and xline
    if all(v is not None for v in (head_loc.iline, head_loc.xline)):
        print("Loading as 3D")
        ds = _3dsegy_loader(
            *common_args,
            **common_kwargs,
            **segyio_kwargs,
        )
        indexer = ["il_index", "xl_index"]
        dims = (
            DimensionKeyField.threed_head
            if offset is None
            else DimensionKeyField.threed_ps_head
        )
        is3d2d = True

    # 2d data
    elif head_loc.cdp is not None:
        print("Loading as 2D")
        ds = _2dsegy_loader(*common_args, **common_kwargs, **segyio_kwargs)
        indexer = ["cdp_index"]
        dims = (
            DimensionKeyField.twod_head
            if offset is None
            else DimensionKeyField.twod_ps_head
        )
        is3d2d = True

    # fallbak to just a 2d array of traces
    else:
        ds = _2dsegy_loader(*common_args, **common_kwargs, **segyio_kwargs)
        indexer = []
        dims = DimensionKeyField.cdp_2d
        is3d2d = False

    indexer = indexer + ["off_index"] if offset is not None else indexer

    ds = _loader_converter_write_headers(
        ds, head_df, indexer, dims, extra_byte_fields, is3d2d=is3d2d
    )

    # ds.seis.get_corner_points()
    return ds


def segy_converter(
    segyfile,
    ncfile,
    cdp=None,
    iline=None,
    xline=None,
    cdpx=None,
    cdpy=None,
    offset=None,
    vert_domain="TWT",
    data_type="AMP",
    ix_crop=None,
    cdp_crop=None,
    xy_crop=None,
    z_crop=None,
    return_geometry=False,
    silent=False,
    extra_byte_fields=None,
    **segyio_kwargs,
):
    """Convert SEG-Y data to NetCDF4 File

    The output ncfile has the following structure
        Dimensions:
            cdp/iline - CDP or Inline axis
            xline - Xline axis
            twt/depth - The vertical axis
            offset - Offset/Angle Axis
        Coordinates:
            iline - The inline numbering
            xline - The xline numbering
            cdp_x - Eastings
            cdp_y - Northings
            cdp - Trace Number for 2d
        Variables
            data - The data volume
        Attributes:
            ns - number of samples vertical
            sample_rate - sample rate in ms/m
            test - text header
            measurement_system : m/ft
            source_file : segy source
            srd : seismic reference datum
            percentiles : data amplitude percentiles
            coord_scalar : from trace headers

    Args:
        segyfile (str): Input segy file path
        ncfile (str): Output SEISNC file path. If none the loaded data will be
            returned in memory as an xarray.Dataset.
        cdp (int, optional): The CDP byte location, usually 21.
        iline (int, optional): Inline byte location, usually 189
        xline (int, optional): Cross-line byte location, usually 193
        cdpx (int, optional): UTMX byte location, usually 181
        cdpy (int, optional): UTMY byte location, usually 185
        offset (int, optional): Offset/angle byte location
        vert_domain (str, optional): Vertical sampling domain. One of ['TWT', 'DEPTH']. Defaults to 'TWT'.
        data_type (str, optional): Data type ['AMP', 'VEL']. Defaults to 'AMP'.
        ix_crop (list, optional): List of minimum and maximum inline and crossline to output.
            Has the form '[min_il, max_il, min_xl, max_xl]'. Ignored for 2D data.
        cdp_crop (list, optional): List of minimum and maximum cmp values to output.
            Has the form '[min_cmp, max_cmp]'. Ignored for 3D data.
        xy_crop (list, optional): List of minimum and maximum cdp_x and cdp_y to output.
            Has the form '[min_x, max_x, min_y, max_y]'. Ignored for 2D data.
        z_crop (list, optional): List of minimum and maximum vertical samples to output.
            Has the form '[min, max]'.
        return_geometry (bool, optional): If true returns an xarray.dataset which doesn't contain data but mirrors
            the input volume header information.
        silent (bool): Disable progress bar.
        extra_byte_fields (list/mapping): A list of int or mapping of byte fields that should be returned as variables in the dataset.a
        **segyio_kwargs: Extra keyword arguments for segyio.open

    """
    # Input sanity checks
    extra_byte_fields = _loader_converter_checks(cdp, iline, xline, extra_byte_fields)
    byte_loc = [
        bytel for bytel in [cdp, iline, xline, cdpx, cdpy, offset] if bytel is not None
    ]
    check_tracefield(byte_loc)

    head_df, head_bin, head_loc = _loader_converter_header_handling(
        segyfile,
        cdp=cdp,
        iline=iline,
        xline=xline,
        cdpx=cdpx,
        cdpy=cdpy,
        offset=offset,
        vert_domain=vert_domain,
        data_type=data_type,
        ix_crop=ix_crop,
        cdp_crop=cdp_crop,
        xy_crop=xy_crop,
        z_crop=z_crop,
        return_geometry=return_geometry,
        silent=silent,
        extra_byte_fields=extra_byte_fields,
        optimised_load=True,
        **segyio_kwargs,
    )
    print("header_loaded")
    common_args = (segyfile, head_df, head_bin, head_loc)

    common_kwargs = dict(
        zcrop=z_crop,
        ncfile=ncfile,
        vert_domain=vert_domain,
        data_type=data_type,
        return_geometry=return_geometry,
        silent=silent,
    )

    # 3d data needs iline and xline
    if iline is not None and xline is not None:

        print("is_3d")
        ds = _3dsegy_loader(
            *common_args,
            **common_kwargs,
            **segyio_kwargs,
        )
        indexer = ["il_index", "xl_index"]
        dims = (
            DimensionKeyField.threed_head
            if offset is None
            else DimensionKeyField.threed_ps_head
        )
        is3d2d = True

    # 2d data
    elif cdp is not None:
        ds = _2dsegy_loader(*common_args, **common_kwargs, **segyio_kwargs)
        indexer = ["cdp_index"]
        dims = (
            DimensionKeyField.twod_head
            if offset is None
            else DimensionKeyField.twod_ps_head
        )
        is3d2d = True

    # fallbak to just a 2d array of traces
    else:
        ds = _2dsegy_loader(*common_args, **common_kwargs, **segyio_kwargs)
        indexer = []
        dims = DimensionKeyField.cdp_2d
        is3d2d = False

    indexer = indexer + ["off_index"] if offset is not None else indexer

    ds = _loader_converter_write_headers(
        ds, head_df, indexer, dims, extra_byte_fields, is3d2d=is3d2d
    )
    new_vars = {key: ds[key] for key in extra_byte_fields}
    ds.close()
    del ds

    with h5netcdf.File(ncfile, "a") as seisnc:
        for var, darray in new_vars.items():
            seisnc_var = seisnc.create_variable(var, darray.dims, darray.dtype)
            seisnc_var[...] = darray[...]
            seisnc.flush()

    return None
